# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import math
from spinn_utilities.log import FormatAdapter
from spinn_utilities.config_holder import get_config_float, get_config_int
from spinn_front_end_common.abstract_models.impl.\
    tdma_aware_application_vertex import (
        TDMAAwareApplicationVertex)
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities.constants import CLOCKS_PER_US
logger = FormatAdapter(logging.getLogger(__name__))

# default fraction of the real time we will use for spike transmissions
FRACTION_OF_TIME_FOR_SPIKE_SENDING = 0.8
FRACTION_OF_TIME_STEP_BEFORE_SPIKE_SENDING = 0.1


def local_tdma_builder():
    """
    Builds a localised TDMA.

    Builds a localised TDMA which allows a number of machine vertices
    of the same application vertex to fire at the same time. Ensures that
    other application vertices are not firing at the same time. Verifies if
    the total time required fits into the time scale factor and machine time
    step. Below are text diagrams to show how this works in principle.

    *Figure 1:* bits needed to figure out time between spikes.
    Cores 0-4 have 2 atoms, core 5 has 1 atom::

        #        0     1       2      3       4      5
        # T2-[   X                    X
        #    |         X                      X
        #    |                 X                     X
        #    [  X                     X
        #       |------| T
        #              X                      X
        #                      X <- T3

        T = time_between_cores
        T2 = time_between_phases
        T3 = end of TDMA (equiv of ((n_phases + 1) * T2))
        cutoff = 2. n_phases = 3 max_atoms = 2

    Constants etc just to get into head:

    * clock cycles = 200 MHz = 200 = sv->cpu_clk
    * 1ms = 200000 for timer 1. = clock cycles
    * 200 per microsecond
    * machine time step = microseconds already.
    * `__time_between_cores` = microseconds.

    *Figure 2:* initial offset (used to try to interleave packets from other
    application vertices into the TDMA without extending the overall time, and
    trying to stop multiple packets in flight at same time).

    *Figure 3:* bits needed to figure out time between spikes.
    Cores 0-4 have 2 atoms, core 5 has 1 atom::

        #        0  .5   1   .5    2   .5   3    .5   4   .5   5  .5
        # T2-[   X   Y                      X     Y
        #    |           X   Y                        X    Y
        #    |                     X    Y                      X   Y
        #    [  X    Y                      X     Y
        #       |-------| T
        #               X    Y                        X    Y
        #               |----| T4
        #                   T3 ->  X    Y

        T4 is the spreader between populations.
        X is pop0 firing,
        Y is pop1 firing

    """
    if FecDataView.get_n_vertices() == 0:
        return
    # get config params
    us_per_cycle = FecDataView.get_hardware_time_step_us()
    clocks_per_cycle = us_per_cycle * CLOCKS_PER_US
    (app_machine_quantity, clocks_between_cores, clocks_for_sending,
     clocks_waiting, clocks_initial) = __config_values(clocks_per_cycle)

    # calculate for each app vertex if the time needed fits
    app_verts = list()
    max_fraction_of_sending = 0
    for app_vertex in FecDataView.get_vertices_by_type(
            TDMAAwareApplicationVertex):
        app_verts.append(app_vertex)

        # get timings

        # check config params for better performance
        n_at_same_time, local_clocks = __auto_config_times(
            app_machine_quantity, clocks_between_cores,
            clocks_for_sending, app_vertex, clocks_waiting)
        n_phases, n_slots, clocks_between_phases = __generate_times(
            app_vertex, n_at_same_time, local_clocks)

        # store in tracker
        app_vertex.set_other_timings(
            local_clocks, n_slots, clocks_between_phases,
            n_phases, clocks_per_cycle)

        # test timings
        fraction_of_sending = __get_fraction_of_sending(
            n_phases, clocks_between_phases, clocks_for_sending)
        if fraction_of_sending is not None:
            max_fraction_of_sending = max(
                max_fraction_of_sending, fraction_of_sending)

    time_scale_factor_needed = (
            FecDataView.get_time_scale_factor() * max_fraction_of_sending)
    if max_fraction_of_sending > 1:
        logger.warning(
            "A time scale factor of {} may be needed to run correctly",
            time_scale_factor_needed)

    # get initial offset for each app vertex.
    for app_vertex in FecDataView.get_vertices_by_type(
            TDMAAwareApplicationVertex):
        initial_offset = __generate_initial_offset(
            app_vertex, app_verts, clocks_initial,
            clocks_waiting)
        app_vertex.set_initial_offset(initial_offset)


def __auto_config_times(
        app_machine_quantity, clocks_between_cores, clocks_for_sending,
        app_vertex, clocks_waiting):

    n_cores = app_vertex.get_n_cores()
    n_phases = app_vertex.get_n_phases()

    # If there are no packets sent, pretend there is 1 to avoid division
    # by 0; it won't actually matter anyway
    if n_phases == 0:
        n_phases = 1

    # Overall time of the TDMA window minus initial offset
    overall_clocks_available = clocks_for_sending - clocks_waiting

    # Easier bool compares
    core_set = clocks_between_cores is not None
    app_set = app_machine_quantity is not None

    # Adjust time between cores to fit time scale
    if not core_set and app_set:
        n_slots = int(math.ceil(n_cores / app_machine_quantity))
        clocks_per_phase = (
            int(math.ceil(overall_clocks_available / n_phases)))
        clocks_between_cores = clocks_per_phase / n_slots
        logger.debug(
            "adjusted clocks between cores is {}",
            clocks_between_cores)

    # Adjust cores at same time to fit time between cores.
    if core_set and not app_set:
        clocks_per_phase = (
            int(math.ceil(overall_clocks_available / n_phases)))
        max_slots = int(math.floor(
            clocks_per_phase / clocks_between_cores))
        app_machine_quantity = int(math.ceil(n_cores / max_slots))
        logger.debug(
            "Adjusted the number of cores of a app vertex that "
            "can fire at the same time to {}",
            app_machine_quantity)

    return app_machine_quantity, clocks_between_cores


def __generate_initial_offset(
        app_vertex, app_verts, clocks_between_cores, clocks_waiting):
    """
    Calculates from the app vertex index the initial offset for the
    TDMA between all cores.

    :param ~pacman.model.graphs.application.ApplicationVertex app_vertex:
        the app vertex in question.
    :param app_verts: the list of app vertices.
    :type app_verts:
        list(~pacman.model.graphs.application.ApplicationVertex)
    :param int clocks_between_cores: the clock cycles between cores.
    :param int clocks_waiting: the clock cycles to wait for.
    :return: the initial offset for this app vertex including wait time.
    :rtype: int
    """
    # This is an offset between cores
    initial_offset_clocks = int(math.ceil(
        (app_verts.index(app_vertex) * clocks_between_cores) /
        len(app_verts)))
    # add the offset the clocks to wait for before sending anything at all
    initial_offset_clocks += clocks_waiting
    return initial_offset_clocks


def __generate_times(
        app_vertex, app_machine_quantity, clocks_between_cores):
    """
    Generates the number of phases needed for this app vertex, as well as the
    number of slots and the time between spikes for this app vertex, given the
    number of machine verts to fire at the same time from a given app vertex.

    :param TDMAAwareApplicationVertex app_vertex: the app vertex
    :param int app_machine_quantity: the pop spike control level
    :param int clocks_between_cores: the clock cycles between cores
    :return: (n_phases, n_slots, time_between_phases) for this app vertex
    :rtype: tuple(int, int, int)
    """

    # Figure total T2s
    n_phases = app_vertex.get_n_phases()

    # how many hops between T2's
    n_cores = app_vertex.get_n_cores()
    n_slots = int(math.ceil(n_cores / app_machine_quantity))

    # figure T2
    clocks_between_phases = int(math.ceil(clocks_between_cores * n_slots))

    return n_phases, n_slots, clocks_between_phases


def __get_fraction_of_sending(
        n_phases, clocks_between_phases, clocks_for_sending):
    """
    Get the fraction of the send.

    :param int n_phases:
        the max number of phases this TDMA needs for a given app vertex
    :param int time_between_phases: the time between phases.
    :param float fraction_of_sending:
        fraction of time step for sending packets
    :param str label: the app vertex we're considering at this point
    :rtype: float
    """
    # figure how much time this TDMA needs
    total_clocks_needed = n_phases * clocks_between_phases
    return total_clocks_needed / clocks_for_sending


def __check_at_most_one(name_1, value_1, name_2, value_2):
    if value_1 is not None and value_2 is not None:
        raise ConfigurationException(
            f"Both {name_1} and {name_2} have been specified; "
            "please choose just one")


def __check_only_one(name_1, value_1, name_2, value_2):
    """
    Checks that exactly one of the values is not `None`.
    """
    __check_at_most_one(name_1, value_1, name_2, value_2)
    if value_1 is None and value_2 is None:
        raise ConfigurationException(
            f"Exactly one of {name_1} and {name_2} must be specified")


def __config_values(clocks_per_cycle):
    """
    Read the configuration for the right parameters and combinations.

    :param int clocks_per_cycle: The number of clock cycles per time step
    :return: (app_machine_quantity, clocks_between_cores,
            clocks_for_sending, clocks_waiting, initial_clocks)
    :rtype: tuple(int, int, int, int. int)
    """
    # set the number of cores expected to fire at any given time
    app_machine_quantity = get_config_int(
        "Simulation", "app_machine_quantity")

    # set the time between cores to fire
    time_between_cores = get_config_float(
        "Simulation", "time_between_cores")
    clocks_between_cores = get_config_int(
        "Simulation", "clock_cycles_between_cores")
    __check_at_most_one(
        "time_between_cores", time_between_cores,
        "clock_cycles_betwen_cores", clocks_between_cores)
    if (time_between_cores is None and clocks_between_cores is None and
            app_machine_quantity is None):
        raise ConfigurationException(
            "Either one of time_between_cores and clocks_between_cores"
            " must be specified or else app_machine_quantity must be"
            " specified")
    if time_between_cores is not None:
        clocks_between_cores = time_between_cores * CLOCKS_PER_US

    # time spend sending
    fraction_of_sending = get_config_float(
        "Simulation", "fraction_of_time_spike_sending")
    clocks_for_sending = get_config_int(
        "Simulation", "clock_cycles_sending")
    __check_only_one(
        "fraction_of_time_spike_sending", fraction_of_sending,
        "clock_cycles_sending", clocks_for_sending)
    if fraction_of_sending is not None:
        clocks_for_sending = int(round(
            clocks_per_cycle * fraction_of_sending))

    # time waiting before sending
    fraction_of_waiting = get_config_float(
        "Simulation", "fraction_of_time_before_sending")
    clocks_waiting = get_config_int(
        "Simulation", "clock_cycles_before_sending")
    __check_only_one(
        "fraction_of_time_before_sending", fraction_of_waiting,
        "clock_cycles_before_sending", clocks_waiting)
    if fraction_of_waiting is not None:
        clocks_waiting = int(round(clocks_per_cycle * fraction_of_waiting))

    # time to offset app vertices between each other
    fraction_initial = get_config_float(
        "Simulation", "fraction_of_time_for_offset")
    clocks_initial = get_config_int(
        "Simulation", "clock_cycles_for_offset")
    __check_only_one(
        "fraction_of_time_for_offset", fraction_initial,
        "clock_cycles_for_offset", clocks_initial)
    if fraction_initial is not None:
        clocks_initial = int(round(clocks_per_cycle * fraction_initial))

    # check fractions less than 1.
    if (clocks_for_sending + clocks_waiting + clocks_initial >
            clocks_per_cycle):
        raise ConfigurationException(
            "The total time for the TDMA must not exceed the time per"
            " cycle")

    return (app_machine_quantity, clocks_between_cores,
            clocks_for_sending, clocks_waiting, clocks_initial)



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


class AbstractProvidesLocalProvenanceData(object, metaclass=AbstractBase):
    """
    Indicates an object that provides locally obtained provenance data.

    GraphProvenanceGatherer will check all Vertices and all Edges in
    the ApplicationGraph
    """

    __slots__ = ()

    @abstractmethod
    def get_local_provenance_data(self):
        """
        Get an iterable of provenance data items.

        :rtype: iterable
        """



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from spinnman.model.enums import CPUState


class ExecutableType(Enum):
    """
    The different types of executable from the perspective of how they
    are started and controlled.
    """

    #: Runs immediately without waiting for barrier and then exits.
    RUNNING = (
        0,
        [CPUState.RUNNING],
        [CPUState.FINISHED],
        False,
        "Runs immediately without waiting for barrier and then exits")
    #: Calls ``spin1_start(SYNC_WAIT)`` and then eventually ``spin1_exit()``.
    SYNC = (
        1,
        [CPUState.SYNC0],
        [CPUState.FINISHED],
        False,
        "Calls spin1_start(SYNC_WAIT) and then eventually spin1_exit()")
    #: Calls ``simulation_run()`` and ``simulation_exit()`` /
    #: ``simulation_handle_pause_resume()``.
    USES_SIMULATION_INTERFACE = (
        2,
        [CPUState.SYNC0, CPUState.SYNC1, CPUState.PAUSED, CPUState.READY],
        [CPUState.READY],
        True,
        "Calls simulation_run() and simulation_exit() / "
        "simulation_handle_pause_resume()")
    #: Situation where there user has supplied no application but for some
    #: reason still wants to run.
    NO_APPLICATION = (
        3,
        [],
        [],
        True,
        "Situation where there user has supplied no application but for "
        "some reason still wants to run")
    #: Runs immediately without waiting for barrier and never ends.
    SYSTEM = (
        4,
        [CPUState.RUNNING],
        [CPUState.RUNNING],
        True,
        "Runs immediately without waiting for barrier and never ends")

    def __new__(cls, value, start_state, end_state,
                supports_auto_pause_and_resume, doc=""):
        # pylint: disable=protected-access, too-many-arguments
        obj = object.__new__(cls)
        obj._value_ = value
        obj.start_state = start_state
        obj.end_state = end_state
        obj.supports_auto_pause_and_resume = supports_auto_pause_and_resume
        obj.__doc__ = doc
        return obj

    def __init__(self, value, start_state, end_state,
                 supports_auto_pause_and_resume, doc=""):
        # pylint: disable=too-many-arguments
        self._value_ = value
        self.__doc__ = doc
        self.start_state = start_state
        self.end_state = end_state
        self.supports_auto_pause_and_resume = supports_auto_pause_and_resume
        self.__doc__ = doc



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import AbstractContextManager as ACMBase
import enum
import hashlib
import logging
import os
import pathlib
import sqlite3
import struct
from spinn_utilities.abstract_context_manager import AbstractContextManager
from spinn_utilities.logger_utils import warn_once
logger = logging.getLogger(__name__)


class Isolation(enum.Enum):
    """
    Transaction isolation levels for :py:meth:`SQLiteDB.transaction`.
    """
    #: Standard transaction type; postpones holding a lock until required.
    DEFERRED = "DEFERRED"
    #: Take the lock immediately; this may be a read-lock that gets upgraded.
    IMMEDIATE = "IMMEDIATE"
    #: Take a write lock immediately. This is the strongest lock type.
    EXCLUSIVE = "EXCLUSIVE"


class SQLiteDB(AbstractContextManager):
    """
    General support class for SQLite databases. This handles a lot of the
    low-level detail of setting up a connection.

    Basic usage (with the default row type)::

        with SQLiteDB("db_file.sqlite3") as db:
            with db.transaction() as cursor:
                for row in cursor.execute("SELECT thing FROM ..."):
                    print(row["thing"])

    This class is designed to be either used as above or by subclassing.
    See the `SQLite SQL documentation <https://www.sqlite.org/lang.html>`_ for
    details of how to write queries, and the Python :py:mod:`sqlite3` module
    for how to do parameter binding.
    """

    __slots__ = [
        # the database holding the data to store
        "__db",
    ]

    def __init__(self, database_file=None, *, read_only=False, ddl_file=None,
                 row_factory=sqlite3.Row, text_factory=memoryview,
                 case_insensitive_like=True):
        """
        :param str database_file:
            The name of a file that contains (or will contain) an SQLite
            database holding the data. If omitted, an unshared in-memory
            database will be used (suitable only for testing).
        :param bool read_only:
            Whether the database is in read-only mode. When the database is in
            read-only mode, it *must* already exist.
        :param ddl_file:
            The name of a file (typically containing SQL DDL commands used to
            create the tables) to be evaluated against the database before this
            object completes construction. If ``None``, nothing will be
            evaluated. You probably don't want to specify a DDL file at the
            same time as setting ``read_only=True``.
        :type ddl_file: str or None
        :param row_factory:
            Callable used to create the rows of result sets.
            Either ``tuple`` or ``sqlite3.Row`` (default);
            can be ``None`` to use the DB driver default.
        :type row_factory: ~collections.abc.Callable or None
        :param text_factory:
            Callable used to create the Python values of non-numeric columns
            in result sets. Usually ``memoryview`` (default) but should be
            ``str`` when you're expecting string results;
            can be ``None`` to use the DB driver default.
        :type text_factory: ~collections.abc.Callable or None
        :param bool case_insensitive_like:
            Whether we want the ``LIKE`` matching operator to be case-sensitive
            or case-insensitive (default).
        """
        self.__db = None
        if database_file is None:
            self.__db = sqlite3.connect(":memory:")  # Magic name!
            # in-memory DB is never read-only
        elif read_only:
            if not os.path.exists(database_file):
                raise FileNotFoundError(f"no such DB: {database_file}")
            db_uri = pathlib.Path(os.path.abspath(database_file)).as_uri()
            # https://stackoverflow.com/a/21794758/301832
            self.__db = sqlite3.connect(f"{db_uri}?mode=ro", uri=True)
        else:
            self.__db = sqlite3.connect(database_file)

        if row_factory:
            self.__db.row_factory = row_factory
        if text_factory:
            self.__db.text_factory = text_factory

        if not read_only and ddl_file:
            with open(ddl_file, encoding="utf-8") as f:
                sql = f.read()
            self.__db.executescript(sql)
            # Stamp the DB with a schema version, which is the first four
            # bytes of the MD5 hash of the content of the schema file used to
            # set it up. We don't currently validate this at all.
            #
            # The application_id pragma would be used within the DDL schema.
            ddl_hash, = struct.unpack_from(
                ">I", hashlib.md5(sql.encode()).digest())
            self.pragma("user_version", ddl_hash)
        if case_insensitive_like:
            self.pragma("case_sensitive_like", False)
        # Official recommendations!
        self.pragma("foreign_keys", True)
        self.pragma("recursive_triggers", True)
        self.pragma("trusted_schema", False)

    def __del__(self):
        self.close()

    def close(self):
        """
        Finalises and closes the database.
        """
        try:
            if self.__db is not None:
                self.__db.close()
                self.__db = None
        except AttributeError:
            self.__db = None

    def pragma(self, pragma_name, value):
        """
        Set a database ``PRAGMA``. See the `SQLite PRAGMA documentation
        <https://www.sqlite.org/pragma.html>`_ for details.

        :param str pragma_name:
            The name of the pragma to set.
        :param value:
            The value to set the pragma to.
        :type value: bool or int or str
        """
        if isinstance(value, bool):
            if value:
                self.__db.executescript(f"PRAGMA {pragma_name}=ON;")
            else:
                self.__db.executescript(f"PRAGMA {pragma_name}=OFF;")
        elif isinstance(value, int):
            self.__db.executescript(f"PRAGMA {pragma_name}={value};")
        elif isinstance(value, str):
            self.__db.executescript(f"PRAGMA {pragma_name}='{value}';")
        else:
            raise TypeError("can only set pragmas to bool, int or str")

    @property
    def connection(self):
        """
        The underlying SQLite database connection.

        .. warning::
            If you're using this a lot, consider contacting the SpiNNaker
            Software Team with details of your use case so we can extend the
            relevant core class to support you. *Normally* it is better to use
            :py:meth:`transaction` to obtain a cursor with appropriate
            transactional guards.

        :rtype: ~sqlite3.Connection
        :raises AttributeError: if the database connection has been closed
        """
        warn_once(
            logger,
            "Low-level connection used instead of transaction() context. "
            "Please contact SpiNNaker Software Team with your use-case for "
            "assistance.")
        if not self.__db:
            raise AttributeError("database has been closed")
        return self.__db

    def transaction(self, isolation_level=None):
        """
        Get a context manager that manages a transaction on the database.
        The value of the context manager is a :py:class:`~sqlite3.Cursor`.
        This means you can do this::

            with db.transaction() as cursor:
                cursor.execute(...)

        :param Isolation isolation_level:
            The transaction isolation level.

            .. note::
                This sets it for the connection!
                Can usually be *not* specified.
        :rtype: ~typing.ContextManager(~sqlite3.Cursor)
        """
        if not self.__db:
            raise AttributeError("database has been closed")
        db = self.__db
        if isolation_level:
            db.isolation_level = isolation_level.value
        return _DbWrapper(db)


class _DbWrapper(ACMBase):
    def __init__(self, db):
        self.__d = db

    def __enter__(self):
        self.__d.__enter__()
        return self.__d.cursor()

    def __exit__(self, exc_type, exc_value, traceback):
        return self.__d.__exit__(exc_type, exc_value, traceback)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

from spinn_utilities.log import FormatAdapter
from spinnman.model import ExecutableTargets, CPUInfos
from spinnman.model.enums import CPUState
from spinn_front_end_common.data import FecDataView
from .iobuf_extractor import IOBufExtractor

logger = FormatAdapter(logging.getLogger(__name__))


def _emergency_state_check():
    """
    :param int app_id: the app id
    """
    # pylint: disable=broad-except
    try:
        app_id = FecDataView.get_app_id()
        txrx = FecDataView.get_transceiver()
        rte_count = txrx.get_core_state_count(
            app_id, CPUState.RUN_TIME_EXCEPTION)
        watchdog_count = txrx.get_core_state_count(app_id, CPUState.WATCHDOG)
        if rte_count or watchdog_count:
            states = txrx.get_cores_in_state(
                None, [CPUState.RUN_TIME_EXCEPTION, CPUState.WATCHDOG])
            logger.warning(
                "unexpected core states (rte={}, wdog={})",
                rte_count, watchdog_count)
            logger.warning(txrx.get_core_status_string(states))
    except Exception:
        logger.exception(
            "Could not read the status count - going to individual cores")
        machine = txrx.get_machine_details()
        infos = CPUInfos()
        errors = list()
        for chip in machine.chips:
            for p in chip.processors:
                try:
                    info = txrx.get_cpu_information_from_core(
                        chip.x, chip.y, p)
                    if info.state in (
                            CPUState.RUN_TIME_EXCEPTION, CPUState.WATCHDOG):
                        infos.add_processor(chip.x, chip.y, p, info)
                except Exception:
                    errors.append((chip.x, chip.y, p))
        if len(infos):
            logger.warning(txrx.get_core_status_string(infos))
        if len(len(errors) > 10):
            logger.warning(
                "Could not read information from {} cores", len(errors))
        else:
            logger.warning(
                "Could not read information from cores {}", errors)


def _emergency_iobuf_extract(executable_targets=None):
    """
    :param executable_targets:
        The specific targets to extract, or `None` for all
    :type executable_targets: ExecutableTargets or None
    """
    # pylint: disable=protected-access
    extractor = IOBufExtractor(
        executable_targets,
        recovery_mode=True, filename_template="emergency_iobuf_{}_{}_{}.txt")
    extractor.extract_iobuf()


def emergency_recover_state_from_failure(vertex, placement):
    """
    Used to get at least *some* information out of a core when something
    goes badly wrong. Not a replacement for what abstract spinnaker base does.

    :param ~spinnman.transceiver.Transceiver txrx: The transceiver.
    :param AbstractHasAssociatedBinary vertex:
        The vertex to retrieve the IOBUF from if it is suspected as being dead
    :param ~pacman.model.placements.Placement placement:
        Where the vertex is located.
    """
    # pylint: disable=protected-access
    _emergency_state_check()
    target = ExecutableTargets()
    path = FecDataView.get_executable_path(vertex.get_binary_file_name())
    target.add_processor(
        path, placement.x, placement.y, placement.p,
        vertex.get_binary_start_type())
    _emergency_iobuf_extract(target)


def emergency_recover_states_from_failure():
    """
    Used to get at least *some* information out of a core when something
    goes badly wrong. Not a replacement for what abstract spinnaker base does.

    :param ~spinnman.model.ExecutableTargets executable_targets:
        The what/where mapping
    """
    _emergency_state_check()
    _emergency_iobuf_extract()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import json
import logging
import os
import subprocess
from spinn_utilities.config_holder import get_config_str
from spinn_utilities.log import FormatAdapter
from pacman.exceptions import PacmanExternalAlgorithmFailedToCompleteException
from pacman.model.graphs import AbstractVirtual
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.report_functions import (
    write_json_machine)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.interface.buffer_management.buffer_models import (
    AbstractReceiveBuffersToHost)
from spinn_front_end_common.interface.buffer_management.storage_objects \
    import BufferDatabase
from spinn_front_end_common.interface.ds import DsSqlliteDatabase

logger = FormatAdapter(logging.getLogger(__name__))


class JavaCaller(object):
    """
    Support class that holds all the stuff for running stuff in Java.
    This includes the work of preparing data for transmitting to Java and
    back.

    This separates the choices of how to call the Java batch vs streaming,
    jar locations, parameters, etc. from the rest of the Python code.
    """

    __slots__ = [
        "_chipxy_by_ethernet",
        # The call to get java to work. Including the path if required.
        "_java_call",
        # The location of the java jar file
        "_jar_file",
        # The location where the machine json is written
        "_machine_json_path",
        # Dict of chip (x, y) to the p of the monitor vertex
        "_monitor_cores",
        # Flag to indicate if at least one placement is recording
        "_recording",
        # Dict of ethernet (x, y) and the packetGather IPtago
        "_gatherer_iptags",
        # Dict of ethernet (x, y) to the p of the packetGather vertex
        "_gatherer_cores",
        # The location where the latest placement json is written
        "_placement_json",
        # Properties flag to be passed to Java
        "_java_properties"
    ]

    def __init__(self):
        """
        Creates a Java caller and checks the user/configuration parameters.

        :raise ConfigurationException: if simple parameter checking fails.
        """
        self._recording = None
        self._java_call = get_config_str("Java", "java_call")
        result = subprocess.call([self._java_call, '-version'])
        if result != 0:
            raise ConfigurationException(
                f" {self._java_call} -version failed. "
                "Please set [Java] java_call to the absolute path "
                "to start java. (in config file)")

        self._find_java_jar()

        self._machine_json_path = None
        self._placement_json = None
        self._monitor_cores = None
        self._gatherer_iptags = None
        self._gatherer_cores = None
        self._java_properties = get_config_str("Java", "java_properties")
        self._chipxy_by_ethernet = None
        if self._java_properties is not None:
            self._java_properties = self._java_properties.split()
            # pylint: disable=not-an-iterable
            for _property in self._java_properties:
                if _property[:2] != "-D":
                    raise ConfigurationException(
                        "Java Properties must start with -D "
                        f"found at {_property}")

    def _find_java_jar(self):
        java_spinnaker_path = get_config_str("Java", "java_spinnaker_path")
        java_jar_path = get_config_str("Java", "java_jar_path")
        if java_spinnaker_path is None:
            interface = os.path.dirname(os.path.realpath(__file__))
            spinn_front_end_common = os.path.dirname(interface)
            github_checkout_dir = os.path.dirname(spinn_front_end_common)
            parent = os.path.dirname(github_checkout_dir)
            java_spinnaker_path = os.path.join(parent, "JavaSpiNNaker")
        else:
            # As I don't know how to write pwd and /JavaSpiNNaker to one line
            indirect_path = os.path.join(
                java_spinnaker_path, "JavaSpiNNaker")
            if os.path.isdir(indirect_path):
                java_spinnaker_path = indirect_path
        auto_jar_file = os.path.join(
            java_spinnaker_path, "SpiNNaker-front-end",
            "target", "spinnaker-exe.jar")
        if os.path.exists(auto_jar_file):
            if (java_jar_path is None) or (java_jar_path == auto_jar_file):
                self._jar_file = auto_jar_file
            else:
                raise ConfigurationException(
                    f"Found a jar file at {auto_jar_file} "
                    "while java_jar_path as set. "
                    "Please delete on of the two.")
        else:
            if java_jar_path is None:
                if not os.path.isdir(java_spinnaker_path):
                    raise ConfigurationException(
                        f"No Java code found at {java_spinnaker_path} "
                        "nor is java_jar_path set.")
                else:
                    raise ConfigurationException(
                        f"No jar file at {auto_jar_file} "
                        "nor is java_jar_path set.")
            elif os.path.exists(java_jar_path):
                self._jar_file = auto_jar_file
            else:
                raise ConfigurationException(
                    f"No file found at java_jar_path: {java_jar_path}")

    def set_advanced_monitors(self):
        """
        Create information describing what's going on with the monitor cores.
        """
        tags = FecDataView.get_tags()
        self._monitor_cores = dict()
        for core, monitor_core in FecDataView.iterate_monitor_items():
            placement = FecDataView.get_placement_of_vertex(monitor_core)
            self._monitor_cores[core] = placement.p

        self._gatherer_iptags = dict()
        self._gatherer_cores = dict()
        for core, packet_gather in FecDataView.iterate_gather_items():
            self._gatherer_iptags[core] = \
                tags.get_ip_tags_for_vertex(packet_gather)[0]
            placement = FecDataView.get_placement_of_vertex(packet_gather)
            self._gatherer_cores[core] = placement.p

        self._chipxy_by_ethernet = defaultdict(list)
        machine = FecDataView.get_machine()
        for chip in machine.chips:
            chip_xy = (chip.x, chip.y)
            ethernet = (chip.nearest_ethernet_x, chip.nearest_ethernet_y)
            self._chipxy_by_ethernet[ethernet].append(chip_xy)

    def _machine_json(self):
        """
        Converts the machine in this class to JSON.

        :return: the name of the file containing the JSON
        """
        if self._machine_json_path is None:
            self._machine_json_path = write_json_machine(
                progress_bar=False, validate=False)
        return self._machine_json_path

    def set_placements(self, used_placements):
        """
        Passes in the placements leaving this class to decide pass it to
        Java.

        Currently the extra information extracted is recording region base
        address but this could change if recording region saved in the
        database.

        Currently this method uses JSON but that may well change to using the
        database.

        :param ~pacman.model.placements.Placements used_placements:
            Placements that are recording. May not be all placements
        """
        path = os.path.join(
            FecDataView.get_json_dir_path(), "java_placements.json")
        self._recording = False
        if self._gatherer_iptags is None:
            self._placement_json = self._write_placements(
                used_placements, path)
        else:
            self._placement_json = self._write_gather(
                used_placements, path)

    def _json_placement(self, placement):
        """
        :param ~pacman.model.placements.Placement placement:
        :rtype: dict
        """
        vertex = placement.vertex
        json_placement = dict()
        json_placement["x"] = placement.x
        json_placement["y"] = placement.y
        json_placement["p"] = placement.p

        json_vertex = dict()
        json_vertex["label"] = vertex.label
        if isinstance(vertex, AbstractReceiveBuffersToHost) and \
                vertex.get_recorded_region_ids():
            self._recording = True
            json_vertex["recordedRegionIds"] = vertex.get_recorded_region_ids()
            json_vertex["recordingRegionBaseAddress"] = \
                vertex.get_recording_region_base_address(placement)
        else:
            json_vertex["recordedRegionIds"] = []
            json_vertex["recordingRegionBaseAddress"] = 0
        json_placement["vertex"] = json_vertex

        return json_placement

    def _json_iptag(self, iptag):
        """
        :param ~pacman.model.tags.IPTag iptag:
        :rtype: dict
        """
        json_tag = dict()
        json_tag["x"] = iptag.destination_x
        json_tag["y"] = iptag.destination_y
        json_tag["boardAddress"] = iptag.board_address
        json_tag["targetAddress"] = iptag.ip_address
        # Intentionally not including port!
        json_tag["stripSDP"] = iptag.strip_sdp
        json_tag["tagID"] = iptag.tag
        json_tag["trafficIdentifier"] = iptag.traffic_identifier

        return json_tag

    def _placements_grouped(self, recording_placements):
        """
        :param ~pacman.model.placements.Placements recording_placementss:
        :rtype: dict(tuple(int,int),dict(tuple(int,int),
            ~pacman.model.placements.Placement))
        """
        by_ethernet = defaultdict(lambda: defaultdict(list))
        for placement in recording_placements:
            if not isinstance(placement.vertex, AbstractVirtual):
                machine = FecDataView.get_machine()
                chip = machine.get_chip_at(placement.x, placement.y)
                chip_xy = (placement.x, placement.y)
                ethernet = (chip.nearest_ethernet_x, chip.nearest_ethernet_y)
                by_ethernet[ethernet][chip_xy].append(placement)
        return by_ethernet

    def _write_gather(self, used_placements, path):
        """
        :param ~pacman.model.placements.Placements used_placements:
            placements that are being used. May not be all placements
        :param str path:
        :rtype: str
        """
        placements_by_ethernet = self._placements_grouped(used_placements)
        json_obj = list()
        for ethernet in self._chipxy_by_ethernet:
            by_chip = placements_by_ethernet[ethernet]
            json_gather = dict()
            json_gather["x"] = ethernet[0]
            json_gather["y"] = ethernet[1]
            json_gather["p"] = self._gatherer_cores[ethernet]
            json_gather["iptag"] = self._json_iptag(
                self._gatherer_iptags[ethernet])
            json_chips = list()
            for chip_xy in self._chipxy_by_ethernet[ethernet]:
                json_chip = dict()
                json_chip["x"] = chip_xy[0]
                json_chip["y"] = chip_xy[1]
                json_chip["p"] = self._monitor_cores[chip_xy]
                if chip_xy in by_chip:
                    json_placements = list()
                    for placement in by_chip[chip_xy]:
                        json_p = self._json_placement(placement)
                        if json_p:
                            json_placements.append(json_p)
                    if len(json_placements) > 0:
                        json_chip["placements"] = json_placements
                json_chips.append(json_chip)
            json_gather["monitors"] = json_chips
            json_obj.append(json_gather)

        # dump to json file
        with open(path, "w", encoding="utf-8") as f:
            json.dump(json_obj, f)

        return path

    def _write_placements(self, used_placements, path):
        """
        :param ~pacman.model.placements.Placements placements:
            Placements that are being used. May not be all placements
        :param str path:
        :rtype: str
        """
        # Read back the regions
        json_obj = list()
        for placement in used_placements:
            if not isinstance(placement.vertex, AbstractVirtual):
                json_p = self._json_placement(placement)
                if json_p:
                    json_obj.append(json_p)

        # dump to json file
        with open(path, "w", encoding="utf-8") as f:
            json.dump(json_obj, f)

        return path

    def _run_java(self, *args):
        """
        Does the actual running of `JavaSpiNNaker`. Arguments are those that
        will be processed by the `main` method on the Java side.

        :type list(str) args:
        :rtype: int
        """
        if self._java_properties is None:
            params = [self._java_call, '-jar', self._jar_file]
        else:
            params = [self._java_call] + self._java_properties \
                     + ['-jar', self._jar_file]
        params.extend(args)
        return subprocess.call(params)

    def get_all_data(self):
        """
        Gets all the data from the previously set placements
        and put these in the previously set database.

        :raises PacmanExternalAlgorithmFailedToCompleteException:
            On failure of the Java code.
        """
        if not self._recording:
            return

        if self._gatherer_iptags is None:
            result = self._run_java(
                'download', self._placement_json, self._machine_json(),
                BufferDatabase.default_database_file(),
                FecDataView.get_run_dir_path())
        else:
            result = self._run_java(
                'gather', self._placement_json, self._machine_json(),
                BufferDatabase.default_database_file(),
                FecDataView.get_run_dir_path())
        if result != 0:
            log_file = os.path.join(
                FecDataView.get_run_dir_path(), "jspin.log")
            raise PacmanExternalAlgorithmFailedToCompleteException(
                "Java call exited with value " + str(result) + " see "
                + str(log_file) + " for logged info")

    def execute_data_specification(self):
        """
        Writes all the data specifications, uploading the result to the
        machine.

        :raises PacmanExternalAlgorithmFailedToCompleteException:
            On failure of the Java code.
        """
        result = self._run_java(
            'dse', self._machine_json(),
            DsSqlliteDatabase.default_database_file(),
            FecDataView.get_run_dir_path())
        if result != 0:
            log_file = os.path.join(
                FecDataView.get_run_dir_path(), "jspin.log")
            raise PacmanExternalAlgorithmFailedToCompleteException(
                "Java call exited with value " + str(result) + " see "
                + str(log_file) + " for logged info")

    def execute_system_data_specification(self):
        """
        Writes all the data specifications for system cores,
        uploading the result to the machine.

        :raises PacmanExternalAlgorithmFailedToCompleteException:
            On failure of the Java code.
        """
        result = self._run_java(
            'dse_sys', self._machine_json(),
            DsSqlliteDatabase.default_database_file(),
            FecDataView.get_run_dir_path())
        if result != 0:
            log_file = os.path.join(
                FecDataView.get_run_dir_path(), "jspin.log")
            raise PacmanExternalAlgorithmFailedToCompleteException(
                "Java call exited with value " + str(result) + " see "
                + str(log_file) + " for logged info")

    def execute_app_data_specification(self, use_monitors):
        """
        Writes all the data specifications for application cores,
        uploading the result to the machine.

        .. note::
            May assume that system cores are already loaded and running if
            `use_monitors` is set to `True`.

        :param bool use_monitors:
        :raises PacmanExternalAlgorithmFailedToCompleteException:
            On failure of the Java code.
        """
        if use_monitors:
            result = self._run_java(
                'dse_app_mon', self._placement_json, self._machine_json(),
                DsSqlliteDatabase.default_database_file(),
                FecDataView.get_run_dir_path())
        else:
            result = self._run_java(
                'dse_app', self._machine_json(),
                DsSqlliteDatabase.default_database_file(),
                FecDataView.get_run_dir_path())
        if result != 0:
            log_file = os.path.join(
                FecDataView.get_run_dir_path(), "jspin.log")
            raise PacmanExternalAlgorithmFailedToCompleteException(
                "Java call exited with value " + str(result) + " see "
                + str(log_file) + " for logged info")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import re
from spinn_utilities.log import FormatAdapter
from spinn_utilities.make_tools.replacer import Replacer
from spinn_utilities.progress_bar import ProgressBar
from spinn_machine.core_subsets import CoreSubsets
from spinnman.model.io_buffer import IOBuffer
from spinn_utilities.config_holder import get_config_str
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.helpful_functions import (
    convert_string_into_chip_and_core_subset)


logger = FormatAdapter(logging.getLogger(__name__))
ERROR_ENTRY = re.compile(r"\[ERROR\]\s+\((.*)\):\s+(.*)")
WARNING_ENTRY = re.compile(r"\[WARNING\]\s+\((.*)\):\s+(.*)")
ENTRY_FILE = 1
ENTRY_TEXT = 2


class _DummyProgress(object):
    def over(self, values):
        return values


class IOBufExtractor(object):
    """
    Extract the logging output buffers from the machine, and separates
    lines based on their prefix.
    """

    __slots__ = ["_filename_template", "_recovery_mode", "__system_binaries",
                 "__app_path", "__sys_path",
                 "__suppress_progress", "__from_cores", "__binary_types",
                 "__executable_targets"]

    def __init__(self, executable_targets=None,
                 recovery_mode=False,
                 filename_template="iobuf_for_chip_{}_{}_processor_id_{}.txt",
                 suppress_progress=False):
        """
        :param bool recovery_mode:
        :param str filename_template:
        :param bool suppress_progress:
        :param executable_targets:
            Which Binaries and core to extract from.
            `None` to extract from all.
        :type executable_targets: ~spinnman.model.ExecutableTargets or None
        :param str from_cores:
        :param str binary_types:
        """
        self._filename_template = filename_template
        self._recovery_mode = bool(recovery_mode)
        self.__suppress_progress = bool(suppress_progress)

        self.__app_path = FecDataView.get_app_provenance_dir_path()
        self.__sys_path = FecDataView.get_system_provenance_dir_path()
        self.__from_cores = get_config_str(
            "Reports", "extract_iobuf_from_cores")
        self.__binary_types = get_config_str(
            "Reports", "extract_iobuf_from_binary_types")
        if executable_targets is None:
            self.__executable_targets = FecDataView.get_executable_targets()
        else:
            self.__executable_targets = executable_targets

        self.__system_binaries = set()
        try:
            self.__system_binaries.update(
                self.__executable_targets.get_binaries_of_executable_type(
                    ExecutableType.SYSTEM))
        except KeyError:
            pass

    def extract_iobuf(self):
        """
        Perform the extraction of IOBUF.

        :return: error_entries, warn_entries
        :rtype: tuple(list(str),list(str))
        """
        if self.__from_cores == "ALL":
            return self.__extract_all_cores()
        elif self.__from_cores and self.__binary_types:
            return self.__extract_selected_cores_and_types()
        elif self.__from_cores:
            return self.__extract_selected_cores()
        elif self.__binary_types:
            return self.__extract_selected_types()
        else:
            # nothing
            return [], []

    def __progress(self, bins):
        """
        :param list bins:
        :rtype: ~.ProgressBar
        """
        if self.__suppress_progress:
            return _DummyProgress()
        label = (("Recovering" if self._recovery_mode else "Extracting")
                 + " IOBUF from the machine")
        return ProgressBar(len(bins), label)

    def __prov_path(self, binary):
        """
        :param str binary:
        :return: provenance directory path
        :rtype: str
        """
        return (self.__sys_path if binary in self.__system_binaries
                else self.__app_path)

    def __extract_all_cores(self):
        """
        :rtype: tuple(list(str), list(str))
        """
        error_entries = list()
        warn_entries = list()
        # all the cores
        progress = self.__progress(self.__executable_targets.binaries)
        for binary in progress.over(self.__executable_targets.binaries):
            core_subsets = self.__executable_targets.get_cores_for_binary(
                binary)
            self.__extract_iobufs_for_binary(
                core_subsets, binary, error_entries, warn_entries)
        return error_entries, warn_entries

    def __extract_selected_cores_and_types(self):
        """
        :rtype: tuple(list(str), list(str))
        """
        error_entries = list()
        warn_entries = list()
        # bit of both
        progress = self.__progress(self.__executable_targets.binaries)
        binaries = FecDataView.get_executable_paths(self.__binary_types)
        iocores = convert_string_into_chip_and_core_subset(self.__from_cores)
        for binary in progress.over(self.__executable_targets.binaries):
            if binary in binaries:
                core_subsets = self.__executable_targets.get_cores_for_binary(
                    binary)
            else:
                core_subsets = iocores.intersect(
                    self.__executable_targets.get_cores_for_binary(binary))
            if core_subsets:
                self.__extract_iobufs_for_binary(
                    core_subsets, binary, error_entries, warn_entries)
        return error_entries, warn_entries

    def __extract_selected_cores(self):
        """
        :rtype: tuple(list(str), list(str))
        """
        error_entries = list()
        warn_entries = list()
        # some hard coded cores
        progress = self.__progress(self.__executable_targets.binaries)
        iocores = convert_string_into_chip_and_core_subset(self.__from_cores)
        for binary in progress.over(self.__executable_targets.binaries):
            core_subsets = iocores.intersect(
                self.__executable_targets.get_cores_for_binary(binary))
            if core_subsets:
                self.__extract_iobufs_for_binary(
                    core_subsets, binary, error_entries, warn_entries)
        return error_entries, warn_entries

    def __extract_selected_types(self):
        """
        :rtype: tuple(list(str), list(str))
        """
        error_entries = list()
        warn_entries = list()
        # some binaries
        binaries = FecDataView.get_executable_paths(self.__binary_types)
        progress = self.__progress(binaries)
        for binary in progress.over(binaries):
            core_subsets = self.__executable_targets.get_cores_for_binary(
                binary)
            if core_subsets is not None:
                self.__extract_iobufs_for_binary(
                    core_subsets, binary, error_entries, warn_entries)
        return error_entries, warn_entries

    def __extract_iobufs_for_binary(
            self, core_subsets, binary, error_entries, warn_entries):
        """
        :param ~.CoreSubsets core_subsets: Where the binary is deployed
        :param str binary: What binary was deployed there.
            This is used to determine how to decompress the IOBUF output.
        :param list(str) error_entries:
        :param list(str) warn_entries:
        """
        replacer = Replacer()
        prov_path = self.__prov_path(binary)

        # extract iobuf
        if self._recovery_mode:
            io_buffers = self.__recover_iobufs(core_subsets)
        else:
            io_buffers = list(FecDataView.get_transceiver().get_iobuf(
                core_subsets))

        # write iobuf
        for iobuf in io_buffers:
            self.__process_one_iobuf(iobuf, prov_path, replacer,
                                     error_entries, warn_entries)

    def __process_one_iobuf(
            self, iobuf, file_path, replacer, error_entries, warn_entries):
        """
        :param ~.IOBuffer iobuf:
        :param str file_path:
        :param ~.Replacer replacer:
        :param list(str) error_entries:
        :param list(str) warn_entries:
        """
        file_name = os.path.join(
            file_path, self._filename_template.format(
                iobuf.x, iobuf.y, iobuf.p))

        # set mode of the file based off if the file already exists
        mode = "a" if os.path.exists(file_name) else "w"

        # write iobuf to file and call out errors and warnings.
        with open(file_name, mode, encoding="utf-8") as f:
            for line in iobuf.iobuf.split("\n"):
                replaced = replacer.replace(line)
                f.write(replaced)
                f.write("\n")
                self.__add_value_if_match(
                    ERROR_ENTRY, replaced, error_entries, iobuf)
                self.__add_value_if_match(
                    WARNING_ENTRY, replaced, warn_entries, iobuf)

    def __recover_iobufs(self, core_subsets):
        """
        :param ~.CoreSubsets core_subsets:
        :rtype: list(~.IOBuffer)
        """
        io_buffers = []
        for core_subset in core_subsets:
            for p in core_subset.processor_ids:
                cs = CoreSubsets()
                cs.add_processor(core_subset.x, core_subset.y, p)
                try:
                    transceiver = FecDataView.get_transceiver()
                    io_buffers.extend(transceiver.get_iobuf(cs))
                except Exception as e:  # pylint: disable=broad-except
                    io_buffers.append(IOBuffer(
                        core_subset.x, core_subset.y, p,
                        "failed to retrieve iobufs from "
                        f"{core_subset.x},{core_subset.y},{p}; "
                        f"{str(e)}"))
        return io_buffers

    @staticmethod
    def __add_value_if_match(regex, line, entries, iobuf):
        """
        :param ~typing.Pattern regex:
        :param str line:
        :param list(str) entries:
        :param ~.IOBuffer iobuf:
        """
        match = regex.match(line)
        if match:
            entries.append(f"{iobuf.x}, {iobuf.y}, {iobuf.p}: "
                           f"{match.group(ENTRY_TEXT)} "
                           f"({match.group(ENTRY_FILE)})")



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

from spinn_utilities.log import FormatAdapter
from pacman.data import PacmanDataView
from pacman.model.partitioner_splitters import (
    SplitterOneAppOneMachine, SplitterFixedLegacy)
from pacman.model.graphs.application.abstract import (
    AbstractOneAppOneMachineVertex)
from spinn_front_end_common.utility_models import ReverseIpTagMultiCastSource

logger = FormatAdapter(logging.getLogger(__name__))


def splitter_selector():
    """
    Basic selector which puts the legacy splitter object on
    everything without a splitter object.
    """
    for app_vertex in PacmanDataView.iterate_vertices():
        if app_vertex.splitter is None:
            vertex_selector(app_vertex)


def vertex_selector(app_vertex):
    """
    Main point for selecting a splitter object for a given app vertex.

    Will assume the SplitterFixedLegacy if no heuristic is known for the
    app vertex.

    :param ~pacman.model.graphs.application.ApplicationVertex app_vertex:
        app vertex to give a splitter object to
    """
    if isinstance(app_vertex, AbstractOneAppOneMachineVertex):
        app_vertex.splitter = SplitterOneAppOneMachine()
    elif isinstance(app_vertex, ReverseIpTagMultiCastSource):
        app_vertex.splitter = SplitterFixedLegacy()
    else:
        logger.warning(
            f"The SplitterSelector has not seen the {app_vertex} vertex "
            f"before. Therefore there is no known splitter to allocate to "
            f"this app vertex and so will use the SplitterFixedLegacy.")
        app_vertex.splitter = SplitterFixedLegacy()



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.machine import MachineVertex


@require_subclass(MachineVertex)
class AbstractSendsBuffersFromHost(object, metaclass=AbstractBase):
    """
    A vertex that sends buffers of keys to be
    transmitted at given timestamps in the simulation.
    """

    __slots__ = ()

    @abstractmethod
    def buffering_input(self):
        """
        Return True if the input of this vertex is to be buffered.

        :rtype: bool
        """

    @abstractmethod
    def get_regions(self):
        """
        Get the set of regions for which there are keys to be sent.

        :return: Iterable of region IDs
        :rtype: iterable(int)
        """

    @abstractmethod
    def get_region_buffer_size(self, region):
        """
        Get the size of the buffer to be used in SDRAM on the machine
        for the region in bytes.

        :param int region: The region to get the buffer size of
        :return: The size of the buffer space in bytes
        :rtype: int
        """

    @abstractmethod
    def is_next_timestamp(self, region):
        """
        Determine if there is another timestamp with data to be sent.

        :param int region: The region to determine if there is more data for
        :return: Whether there is more data
        :rtype: bool
        """

    @abstractmethod
    def get_next_timestamp(self, region):
        """
        Get the next timestamp at which there are still keys to be sent
        for the given region.

        :param int region: The region to get the timestamp for
        :return: The timestamp of the next available keys
        :rtype: int
        """

    @abstractmethod
    def is_next_key(self, region, timestamp):
        """
        Determine if there are still keys to be sent at the given
        timestamp for the given region.

        :param int region: The region to determine if there are keys for
        :param int timestamp:
            The timestamp to determine if there are more keys for
        :return: Whether there are more keys to send for the parameters
        :rtype: bool
        """

    @abstractmethod
    def get_next_key(self, region):
        """
        Get the next key in the given region.

        :param int region: The region to get the next key from
        :return: The next key, or `None` if there are no more keys
        :rtype: int
        """

    @abstractmethod
    def is_empty(self, region):
        """
        Return true if there are no spikes to be buffered for the
        specified region.

        :param int region: The region to get the next key from
        :return: Whether there are no keys to send for the region
        :rtype: bool
        """

    @abstractmethod
    def rewind(self, region):
        """
        Rewinds the internal buffer in preparation of re-sending the spikes.

        :param int region: The region to rewind
        """



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .clear_reinjection_queue_message import ClearReinjectionQueueMessage
from .get_reinjection_status_message import (
    GetReinjectionStatusMessage, GetReinjectionStatusMessageResponse)
from .load_application_mc_routes_message import LoadApplicationMCRoutesMessage
from .load_system_mc_routes_message import LoadSystemMCRoutesMessage
from .reset_counters_message import ResetCountersMessage
from .set_reinjection_packet_types_message import (
    SetReinjectionPacketTypesMessage)
from .set_router_timeout_message import SetRouterTimeoutMessage

__all__ = [
    "ClearReinjectionQueueMessage",
    "GetReinjectionStatusMessage",
    "GetReinjectionStatusMessageResponse",
    "LoadApplicationMCRoutesMessage",
    "LoadSystemMCRoutesMessage",
    "ResetCountersMessage",
    "SetReinjectionPacketTypesMessage",
    "SetRouterTimeoutMessage",
    ]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version__ = "1!7.0.0.a4"
__version_month__ = "April"
__version_year__ = "2023"
__version_day__ = "17"
__version_name__ = "Revisionist"



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_front_end_common.utilities.constants import (
    SARK_PER_MALLOC_SDRAM_USAGE, BYTES_PER_WORD)

# Size of data in the header for each recording region
# 1 word for space, 1 word for size+missing, 1 word for pointer
_PER_REGION_HEADER_SIZE = BYTES_PER_WORD * 3


def get_recording_header_size(n_recording_regions):
    """
    Get the size of the data to be written for the recording header.

    This is the data that sets up how recording will be done, and indicates the
    sizes of the regions to be stored.

    :param int n_recording_regions: The number of regions that can be recorded
    :rtype: int
    """
    # See recording.h/recording_initialise for data included in the header
    return BYTES_PER_WORD + (n_recording_regions * _PER_REGION_HEADER_SIZE)


def get_recording_data_constant_size(n_recording_regions):
    """
    Get the size of the headers that are stored in the SDRAM spaces
    allocated during recording_initialise, and so do not need to be
    reserved with DSG (but need to be accounted for in SDRAM calculations).

    :param int n_recording_regions: The number of regions that can be recorded
    :rtype: int
    """
    return (n_recording_regions * SARK_PER_MALLOC_SDRAM_USAGE)


def get_recording_header_array(recorded_region_sizes):
    """
    Get data to be written for the recording header.

    :param list(int) recorded_region_sizes:
        A list of sizes of each region to be recorded.
        A size of 0 is acceptable.
    :rtype: list(int)
    """

    # See recording.h/recording_initialise for data included in the header
    data = list()

    # The parameters
    data.append(len(recorded_region_sizes))

    # Add (space, size, pointer) for each region
    for space in recorded_region_sizes:
        data.extend([space, 0, 0])

    return data



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class RouterSummary(object):

    __slots__ = [
        "_total_entries",
        "_max_per_chip",
        "_max_defaultable",
        "_max_link",
        "_unqiue_routes",
    ]

    def __init__(self, total_entries, max_per_chip, max_defaultable, max_link,
                 unqiue_routes):

        self._total_entries = total_entries
        self._max_per_chip = max_per_chip
        self._max_defaultable = max_defaultable
        self._max_link = max_link
        self._unqiue_routes = unqiue_routes

    @property
    def total_entries(self):
        """
        :rtype: int
        """
        return self._total_entries

    @property
    def max_per_chip(self):
        """
        :rtype: int
        """
        return self._max_per_chip

    @property
    def max_defaultable(self):
        """
        :rtype: int
        """
        return self._max_defaultable

    @property
    def max_link(self):
        """
        :rtype: int
        """
        return self._max_link

    @property
    def unqiue_routes(self):
        """
        :rtype: int
        """
        return self._unqiue_routes



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.application import ApplicationVertex


@require_subclass(ApplicationVertex)
class HasCustomAtomKeyMap(object, metaclass=AbstractBase):
    """
    An object that can provide a custom atom-key mapping for a partition.
    Useful when there isn't a one-to-one correspondence between atoms
    and keys for a given partition.
    """

    @abstractmethod
    def get_atom_key_map(self, pre_vertex, partition_id, routing_info):
        """
        Get the mapping between atoms and keys for the given partition id,
        and for the given machine pre-vertex.

        :param ~pacman.model.graphs.machine.MachineVertex pre_vertex:
            The machine vertex to get the map for
        :param str partition_id: The partition to get the map for
        :param ~pacman.model.routing_info.RoutingInfo routing_info:
            Routing information
        :return: A list of (atom_id, key)
        :rtype: list(tuple(int,int))
        """



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from concurrent.futures import ThreadPoolExecutor, wait  # @UnresolvedImport
from spinn_utilities.abstract_context_manager import AbstractContextManager
from spinn_utilities.config_holder import get_config_bool, get_config_int
from spinn_utilities.log import FormatAdapter
from spinnman.connections.udp_packet_connections import EIEIOConnection
from spinnman.messages.eieio.command_messages import (
    NotificationProtocolDatabaseLocation, NotificationProtocolPauseStop,
    NotificationProtocolStartResume)
from spinnman.exceptions import SpinnmanTimeoutException
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.constants import (
    MAX_DATABASE_PATH_LENGTH)
from spinn_front_end_common.utilities.exceptions import ConfigurationException

logger = FormatAdapter(logging.getLogger(__name__))


class NotificationProtocol(AbstractContextManager):
    """
    The protocol which hand shakes with external devices about the
    database and starting execution.

    The messages sent by this are received by instances of
    :py:class:`DatabaseConnection` (and its subclasses). They are not routed
    via SpiNNaker.
    """
    __slots__ = [
        "__database_message_connections",
        "__sent_visualisation_confirmation",
        "__wait_for_read_confirmation",
        "__wait_for_read_timeout",
        "__wait_futures",
        "__wait_pool"]

    def __init__(self):
        # Determines whether to wait for confirmation that the database
        # has been read before starting the simulation
        self.__wait_for_read_confirmation = get_config_bool(
            "Database", "wait_on_confirmation")
        self.__wait_for_read_timeout = get_config_int(
            "Database", "wait_on_confirmation_timeout")
        self.__wait_pool = ThreadPoolExecutor(max_workers=1)
        self.__wait_futures = list()
        self.__sent_visualisation_confirmation = False
        # These connections are not used to talk to SpiNNaker boards
        # but rather to code running on the current host computer
        self.__database_message_connections = [
            EIEIOConnection(
                local_port=socket_address.listen_port,
                remote_host=socket_address.notify_host_name,
                remote_port=socket_address.notify_port_no)
            for socket_address in
            FecDataView.iterate_database_socket_addresses()]

    def wait_for_confirmation(self):
        """
        If asked to wait for confirmation, waits for all external systems
        to confirm that they are configured and have read the database.
        """
        if self.__wait_for_read_confirmation:
            logger.info("** Awaiting for a response from an external source "
                        "to state its ready for the simulation to start **")
            results = wait(self.__wait_futures,
                           timeout=self.__wait_for_read_timeout)
            if results.not_done:
                raise SpinnmanTimeoutException(
                    f"waiting for external sources: {results.not_done}",
                    self.__wait_for_read_timeout)
        self.__wait_futures = list()

    def send_start_resume_notification(self):
        """
        Either waits till all sources have confirmed read the database
        and are configured, and/or just sends the start notification
        (when the system is executing).
        """
        logger.info("** Sending start / resume message to external sources "
                    "to state the simulation has started or resumed. **")
        if self.__wait_for_read_confirmation:
            self.wait_for_confirmation()
        eieio_command_message = NotificationProtocolStartResume()
        for c in self.__database_message_connections:
            try:
                c.send_eieio_message(eieio_command_message)
            except Exception:  # pylint: disable=broad-except
                logger.warning(
                    "*** Failed to send start/resume notification to external "
                    "application on {}:{} about the simulation ***",
                    c.remote_ip_address, c.remote_port, exc_info=True)

    def send_stop_pause_notification(self):
        """
        Sends the pause / stop notifications when the script has either
        finished or paused.
        """
        logger.info("** Sending pause / stop message to external sources "
                    "to state the simulation has been paused or stopped. **")
        eieio_command_message = NotificationProtocolPauseStop()
        for c in self.__database_message_connections:
            try:
                c.send_eieio_message(eieio_command_message)
            except Exception:  # pylint: disable=broad-except
                logger.warning(
                    "*** Failed to send stop/pause notification to external "
                    "application on {}:{} about the simulation ***",
                    c.remote_ip_address, c.remote_port, exc_info=True)

    # noinspection PyPep8
    def send_read_notification(self):
        """
        Sends notifications to all devices which have expressed an
        interest in when the database has been written
        """
        notification_task = self.__wait_pool.submit(
            self._send_read_notification)
        if self.__wait_for_read_confirmation:
            self.__wait_futures.append(notification_task)

    def _send_read_notification(self):
        """
        Sends notifications to a list of socket addresses that the
        database has been written. Message also includes the path to the
        database

        :param str database_path: the path to the database
        """
        # noinspection PyBroadException
        try:
            self.__do_read_notify()
        except Exception:  # pylint: disable=broad-except
            logger.warning("problem when sending DB notification",
                           exc_info=True)

    def __do_read_notify(self):
        database_path = FecDataView.get_database_file_path()
        # add file path to database into command message.
        if (database_path is not None and
                len(database_path) > MAX_DATABASE_PATH_LENGTH):
            raise ConfigurationException(
                "The file path to the database is too large to be transmitted "
                "via the command packet, please set the file path manually "
                "and set the .cfg parameter [Database] send_file_path to "
                "False")
        message = NotificationProtocolDatabaseLocation(database_path)

        # Send command and wait for response
        logger.info(
            "** Notifying external sources that the database is ready for "
            "reading **")

        # noinspection PyBroadException
        for c in self.__database_message_connections:
            try:
                c.send_eieio_message(message)
            except Exception:  # pylint: disable=broad-except
                logger.warning(
                    "*** Failed to notify external application on {}:{} "
                    "about the database ***",
                    c.remote_ip_address, c.remote_port, exc_info=True)

        self.__sent_visualisation_confirmation = True

        # if the system needs to wait, try receiving a packet back
        if self.__wait_for_read_confirmation:
            for c in self.__database_message_connections:
                try:
                    c.receive_eieio_message()
                    logger.info(
                        "** Confirmation from {}:{} received, continuing **",
                        c.remote_ip_address, c.remote_port)
                except Exception:  # pylint: disable=broad-except
                    logger.warning(
                        "*** Failed to receive notification from external "
                        "application on {}:{} about the database ***",
                        c.remote_ip_address, c.remote_port, exc_info=True)

    @property
    def sent_visualisation_confirmation(self):
        """
        Whether the external application has actually been notified yet.

        :rtype: bool
        """
        return self.__sent_visualisation_confirmation

    def close(self):
        """
        Closes the thread pool and the connections.
        """
        if self.__wait_pool:
            self.__wait_pool.shutdown()
            self.__wait_futures = list()
            self.__wait_pool = None
        for c in self.__database_message_connections:
            c.close()



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.sqlite_db import SQLiteDB
from spinn_front_end_common.abstract_models import (
    AbstractSupportsDatabaseInjection, HasCustomAtomKeyMap)
from spinnman.spalloc import SpallocJob
from spinn_front_end_common.utility_models import LivePacketGather
from pacman.utilities.utility_calls import get_field_based_keys

logger = FormatAdapter(logging.getLogger(__name__))
DB_NAME = "input_output_database.sqlite3"
INIT_SQL = "db.sql"


def _extract_int(x):
    return None if x is None else int(x)


class DatabaseWriter(SQLiteDB):
    """
    The interface for the database system for main front ends.
    Any special tables needed from a front end should be done
    by subclasses of this interface.
    """

    __slots__ = [
        # the path of the database
        "_database_path",

        # the identifier for the SpiNNaker machine
        "_machine_id",

        # Mappings used to accelerate inserts
        "__machine_to_id", "__vertex_to_id"
    ]

    def __init__(self):
        self._database_path = os.path.join(FecDataView.get_run_dir_path(),
                                           DB_NAME)
        init_sql_path = os.path.join(os.path.dirname(__file__), INIT_SQL)

        # delete any old database
        if os.path.isfile(self._database_path):
            os.remove(self._database_path)

        super().__init__(self._database_path, ddl_file=init_sql_path)
        self.__machine_to_id = dict()
        self.__vertex_to_id = dict()

        # set up checks
        self._machine_id = 0

    @staticmethod
    def auto_detect_database():
        """
        Auto detects if there is a need to activate the database system.

        :return: whether the database is needed for the application
        :rtype: bool
        """
        if FecDataView.get_vertices_by_type(LivePacketGather):
            return True
        for vertex in FecDataView.get_vertices_by_type(
                AbstractSupportsDatabaseInjection):
            if vertex.is_in_injection_mode:
                return True
        return False

    @property
    def database_path(self):
        """
        :rtype: str
        """
        return self._database_path

    def __insert(self, cur, sql, *args):
        """
        :param ~sqlite3.Cursor cur:
        :param str sql:
        :rtype: int
        """
        try:
            cur.execute(sql, args)
            return cur.lastrowid
        except Exception:
            logger.exception("problem with insertion; argument types are {}",
                             str(map(type, args)))
            raise

    def add_machine_objects(self):
        """
        Store the machine object into the database.
        """
        machine = FecDataView.get_machine()
        with self.transaction() as cur:
            self.__machine_to_id[machine] = self._machine_id = self.__insert(
                cur,
                """
                INSERT INTO Machine_layout(
                    x_dimension, y_dimension)
                VALUES(?, ?)
                """, machine.width, machine.height)
            cur.executemany(
                """
                INSERT INTO Machine_chip(
                    no_processors, chip_x, chip_y, machine_id,
                    ip_address, nearest_ethernet_x, nearest_ethernet_y)
                VALUES (?, ?, ?, ?, ?, ?, ?)
                """, (
                    (chip.n_processors, chip.x, chip.y, self._machine_id,
                     chip.ip_address,
                     chip.nearest_ethernet_x, chip.nearest_ethernet_y)
                    for chip in machine.chips))

    def add_application_vertices(self):
        """
        Stores the main application graph description (vertices, edges).
        """
        with self.transaction() as cur:
            # add vertices
            for vertex in FecDataView.iterate_vertices():
                vertex_id = self.__insert(
                    cur,
                    "INSERT INTO Application_vertices(vertex_label) VALUES(?)",
                    vertex.label)
                self.__vertex_to_id[vertex] = vertex_id
                for m_vertex in vertex.machine_vertices:
                    m_vertex_id = self.__add_machine_vertex(cur, m_vertex)
                    self.__insert(
                        cur,
                        """
                        INSERT INTO graph_mapper_vertex (
                            application_vertex_id, machine_vertex_id)
                        VALUES(?, ?)
                        """,
                        vertex_id, m_vertex_id)

    def __add_machine_vertex(self, cur, m_vertex):
        m_vertex_id = self.__insert(
            cur, "INSERT INTO Machine_vertices (label)  VALUES(?)",
            str(m_vertex.label))
        self.__vertex_to_id[m_vertex] = m_vertex_id
        return m_vertex_id

    def add_system_params(self, runtime):
        """
        Write system parameters into the database.

        :param int runtime: the amount of time the application is to run for
        """
        with self.transaction() as cur:
            cur.executemany(
                """
                INSERT INTO configuration_parameters (
                    parameter_id, value)
                VALUES (?, ?)
                """, [
                    ("machine_time_step",
                     FecDataView.get_simulation_time_step_us()),
                    ("time_scale_factor",
                     FecDataView.get_time_scale_factor()),
                    ("infinite_run", str(runtime is None)),
                    ("runtime", -1 if runtime is None else runtime),
                    ("app_id", FecDataView.get_app_id())])

    def add_proxy_configuration(self):
        """
        Store the proxy configuration.
        """
        # pylint: disable=protected-access
        if not FecDataView.has_allocation_controller():
            return
        mac = FecDataView.get_allocation_controller()
        if mac.proxying:
            # This is now assumed to be a SpallocJobController;
            # can't check that because of import circularity.
            job = mac._job
            if isinstance(job, SpallocJob):
                with self.transaction() as cur:
                    job._write_session_credentials_to_db(cur)

    def add_placements(self):
        """
        Adds the placements objects into the database.
        """
        with self.transaction() as cur:
            # Make sure machine vertices are represented
            for placement in FecDataView.iterate_placemements():
                if placement.vertex not in self.__vertex_to_id:
                    self.__add_machine_vertex(cur, placement.vertex)
            # add records
            cur.executemany(
                """
                INSERT INTO Placements(
                    vertex_id, chip_x, chip_y, chip_p, machine_id)
                VALUES(?, ?, ?, ?, ?)
                """, (
                    (self.__vertex_to_id[placement.vertex],
                     placement.x, placement.y, placement.p, self._machine_id)
                    for placement in FecDataView.iterate_placemements()))

    def add_tags(self):
        """
        Adds the tags into the database.
        """
        tags = FecDataView.get_tags()
        with self.transaction() as cur:
            cur.executemany(
                """
                INSERT INTO IP_tags(
                    vertex_id, tag, board_address, ip_address, port,
                    strip_sdp)
                VALUES (?, ?, ?, ?, ?, ?)
                """, (
                    (self.__vertex_to_id[vert], ipt.tag, ipt.board_address,
                     ipt.ip_address, ipt.port or 0, 1 if ipt.strip_sdp else 0)
                    for ipt, vert in tags.ip_tags_vertices))

    def create_atom_to_event_id_mapping(self, machine_vertices):
        """
        :param machine_vertices:
        :type machine_vertices:
            list(tuple(~pacman.model.graphs.machine.MachineVertex,int))
        """
        routing_infos = FecDataView.get_routing_infos()
        # This could happen if there are no LPGs
        if machine_vertices is None:
            return
        with self.transaction() as cur:
            for (m_vertex, partition_id) in machine_vertices:
                atom_keys = list()
                if isinstance(m_vertex.app_vertex, HasCustomAtomKeyMap):
                    atom_keys = m_vertex.app_vertex.get_atom_key_map(
                        m_vertex, partition_id, routing_infos)
                else:
                    r_info = routing_infos.get_routing_info_from_pre_vertex(
                        m_vertex, partition_id)
                    # r_info could be None if there are no outgoing edges,
                    # at which point there is nothing to do here anyway
                    if r_info is not None:
                        vertex_slice = m_vertex.vertex_slice
                        keys = get_field_based_keys(r_info.key, vertex_slice)
                        start = vertex_slice.lo_atom
                        atom_keys = [(i, k) for i, k in enumerate(keys, start)]
                m_vertex_id = self.__vertex_to_id[m_vertex]
                cur.executemany(
                    """
                    INSERT INTO event_to_atom_mapping(
                        vertex_id, event_id, atom_id)
                    VALUES (?, ?, ?)
                    """, ((m_vertex_id, int(key), i) for i, key in atom_keys)
                )

    def add_lpg_mapping(self):
        """
        Add mapping from machine vertex to LPG machine vertex.

        :return: A list of (source vertex, partition id)
        :rtype: list(~pacman.model.graphs.machine.MachineVertex, str)
        """
        targets = [(m_vertex, part_id, lpg_m_vertex)
                   for vertex in FecDataView.iterate_vertices()
                   if isinstance(vertex, LivePacketGather)
                   for lpg_m_vertex, m_vertex, part_id
                   in vertex.splitter.targeted_lpgs]

        with self.transaction() as cur:
            cur.executemany(
                """
                INSERT INTO m_vertex_to_lpg_vertex(
                    pre_vertex_id, partition_id, post_vertex_id)
                VALUES(?, ?, ?)
                """, ((self.__vertex_to_id[m_vertex], part_id,
                       self.__vertex_to_id[lpg_m_vertex])
                      for m_vertex, part_id, lpg_m_vertex in targets))

        return [(source, part_id) for source, part_id, _target in targets]



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from spinn_utilities.log import FormatAdapter
from spinn_utilities.socket_address import SocketAddress
from spinnman.data import SpiNNManDataView
from spinnman.messages.scp.enums.signal import Signal
from pacman.data import PacmanDataView
from pacman.model.graphs.application import ApplicationEdge
# in code to avoid circular import
# from spinn_front_end_common.utility_models import LivePacketGather
# from spinn_front_end_common.utility_models import CommandSender

logger = FormatAdapter(logging.getLogger(__name__))
# pylint: disable=protected-access


class _FecDataModel(object):
    """
    Singleton data model.

    This class should not be accessed directly please use the DataView and
    DataWriter classes.
    Accessing or editing the data held here directly is *not supported!*

    There may be other DataModel classes which sit next to this one and hold
    additional data. The DataView and DataWriter classes will combine these
    as needed.

    What data is held where and how can change without notice.
    """

    __singleton = None

    __slots__ = [
        # Data values cached
        "_allocation_controller",
        "_buffer_manager",
        "_current_run_timesteps",
        "_data_in_multicast_key_to_chip_map",
        "_data_in_multicast_routing_tables",
        "_database_file_path",
        "_database_socket_addresses",
        "_dsg_targets",
        "_executable_targets",
        "_executable_types",
        "_first_machine_time_step",
        "_fixed_routes",
        "_gatherer_map",
        "_hardware_time_step_ms",
        "_hardware_time_step_us",
        "_ipaddress",
        "_java_caller",
        "_live_packet_recorder_params",
        "_live_output_vertices",
        "_n_boards_required",
        "_n_chips_required",
        "_n_chips_in_graph",
        "_next_sync_signal",
        "_none_labelled_edge_count",
        "_notification_protocol",
        "_max_run_time_steps",
        "_monitor_map",
        "_reset_number",
        "_run_number",
        "_run_step",
        "_simulation_time_step_ms",
        "_simulation_time_step_per_ms",
        "_simulation_time_step_per_s",
        "_simulation_time_step_s",
        "_simulation_time_step_us",
        "_system_multicast_router_timeout_keys",
        "_timestamp_dir_path",
        "_time_scale_factor",
    ]

    def __new__(cls):
        if cls.__singleton:
            return cls.__singleton
        # pylint: disable=protected-access
        obj = object.__new__(cls)
        cls.__singleton = obj
        obj._notification_protocol = None
        obj._clear()
        return obj

    def _clear(self):
        """
        Clears out all data.
        """
        # Can not be cleared during hard reset as previous runs data checked
        self._database_socket_addresses = set()
        self._executable_types = None
        self._hardware_time_step_ms = None
        self._hardware_time_step_us = None
        self._live_packet_recorder_params = None
        self._live_output_vertices = set()
        self._java_caller = None
        self._n_boards_required = None
        self._n_chips_required = None
        self._none_labelled_edge_count = 0
        self._reset_number = 0
        self._run_number = None
        self._simulation_time_step_ms = None
        self._simulation_time_step_per_ms = None
        self._simulation_time_step_per_s = None
        self._simulation_time_step_s = None
        self._simulation_time_step_us = None
        self._time_scale_factor = None
        self._timestamp_dir_path = None
        self._hard_reset()

    def _hard_reset(self):
        """
        Clears out all data that should change after a reset and graph change.
        """
        self._buffer_manager = None
        self._allocation_controller = None
        self._data_in_multicast_key_to_chip_map = None
        self._data_in_multicast_routing_tables = None
        self._database_file_path = None
        self._dsg_targets = None
        self._executable_targets = None
        self._fixed_routes = None
        self._gatherer_map = None
        self._ipaddress = None
        self._n_chips_in_graph = None
        self._next_sync_signal = Signal.SYNC0
        self._notification_protocol = None
        self._max_run_time_steps = None
        self._monitor_map = None
        self._system_multicast_router_timeout_keys = None
        self._soft_reset()
        self._clear_notification_protocol()

    def _soft_reset(self):
        """
        Clears timing and other data that should changed every reset.
        """
        self._current_run_timesteps = 0
        self._first_machine_time_step = 0
        self._run_step = None

    def _clear_notification_protocol(self):
        if self._notification_protocol:
            try:
                self._notification_protocol.close()
            except Exception as ex:  # pylint: disable=broad-except
                logger.exception(
                    f"Error {ex} when closing the notification_protocol "
                    f"ignored")
        self._notification_protocol = None


class FecDataView(PacmanDataView, SpiNNManDataView):
    """
    Adds the extra Methods to the View for spinn_front_end_commom level.

    See :py:class:`~spinn_utilities.data.UtilsDataView` for a more detailed
    description.

    This class is designed to only be used directly within non-PyNN
    repositories as all methods are available to subclasses
    """

    __fec_data = _FecDataModel()

    __slots__ = []

    # current_run_timesteps and first_machine_time_step

    @classmethod
    def get_current_run_timesteps(cls):
        """
        The end of this or the previous do__run loop time in steps.

        Will be zero if not yet run and not yet in the do_run_loop

        Will be `None` if in run forever mode

        :rtype: int or None
        """
        return cls.__fec_data._current_run_timesteps

    @classmethod
    def get_current_run_time_ms(cls):
        """
        The end of this or the previous do__run loop time in ms.

        Syntactic sugar for `current_run_timesteps * simulation_time_step_ms`

        Will be zero if not yet run and not yet in the do_run_loop

        Will be zero if in run forever mode

        :rtype: float
        """
        if cls.__fec_data._current_run_timesteps is None:
            return 0.0
        return (cls.__fec_data._current_run_timesteps *
                cls.get_simulation_time_step_ms())

    # _allocation_controller
    @classmethod
    def has_allocation_controller(cls):
        """
        Reports if an AllocationController object has already been set.

        :return: True if and only if an AllocationController has been added and
            not reset.
        :rtype: bool
        """
        return cls.__fec_data._allocation_controller is not None

    @classmethod
    def get_allocation_controller(cls):
        """
        Returns the allocation controller if known.

        :rtype: AbstractMachineAllocationController
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the buffer manager unavailable
        """
        if cls.__fec_data._allocation_controller is None:
            raise cls._exception("allocation_controller")

        return cls.__fec_data._allocation_controller

    # _buffer_manager
    @classmethod
    def has_buffer_manager(cls):
        """
        Reports if a BufferManager object has already been set.

        :return:
            True if and only if a BufferManager has been added and not reset
        :rtype: bool
        """
        return cls.__fec_data._buffer_manager is not None

    @classmethod
    def get_buffer_manager(cls):
        """
        Returns the buffer manager if known.

        :rtype:
            ~spinn_front_end_common.interface.buffer_management.BufferManager
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the buffer manager unavailable
        """
        if cls.__fec_data._buffer_manager is None:
            raise cls._exception("buffer_manager")

        return cls.__fec_data._buffer_manager

    @classmethod
    def get_first_machine_time_step(cls):
        """
        The start of this or the next do_run loop time in steps.

        Will be `None` if in run forever mode

        :rtype: int or None
        """
        return cls.__fec_data._first_machine_time_step

    # max_run_time_steps methods

    @classmethod
    def get_max_run_time_steps(cls):
        """
        Returns the calculated longest time this or a future run loop could be.

        Mainly used to indicate the number of timesteps the vertex can and
        therefore should reserve memory for

        Guaranteed to be positive int if available

        :rtype: int
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the max run time is currently unavailable
        """
        if cls.__fec_data._max_run_time_steps is None:
            raise cls._exception("max_run_time_steps")

        return cls.__fec_data._max_run_time_steps

    @classmethod
    def has_max_run_time_steps(cls):
        """
        Indicates if max_run_time_steps is currently available.

        :rtype: bool
        """
        return cls.__fec_data._max_run_time_steps is not None

    # simulation_time_step_methods

    @classmethod
    def has_time_step(cls):
        """
        Check if any/all of the time_step values are known.

        True When all simulation/hardware_time_step methods are known
        False when none of the simulation/hardware_time_step values are known.
        There is never a case when some are known and others not

        :rtype: bool
        """
        return cls.__fec_data._simulation_time_step_us is not None

    @classmethod
    def get_simulation_time_step_us(cls):
        """
        The simulation timestep, in microseconds.

        Previously known as "machine_time_step"

        :rtype: int
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the simulation_time_step_us is currently unavailable
        """
        if cls.__fec_data._simulation_time_step_us is None:
            raise cls._exception("simulation_time_step_us")
        return cls.__fec_data._simulation_time_step_us

    @classmethod
    def get_simulation_time_step_s(cls):
        """
        The simulation timestep, in seconds.

        Syntactic sugar for `simulation_time_step() / 1,000,000`.

        :rtype: float
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the simulation_time_step_ms is currently unavailable
        """
        if cls.__fec_data._simulation_time_step_us is None:
            raise cls._exception("simulation_time_step_s")
        return cls.__fec_data._simulation_time_step_s

    @classmethod
    def get_simulation_time_step_ms(cls):
        """
        The simulation time step, in milliseconds.

        Syntactic sugar for `simulation_time_step_us / 1000`.

        :rtype: float
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the simulation_time_step_ms is currently unavailable
        """
        if cls.__fec_data._simulation_time_step_us is None:
            raise cls._exception("simulation_time_step_ms")
        return cls.__fec_data._simulation_time_step_ms

    @classmethod
    def get_simulation_time_step_per_ms(cls):
        """
        The number of simulation time steps in a millisecond.

        Syntactic sugar for `1000 / simulation_time_step_us`

        :rtype: float
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the simulation_time_step is currently unavailable
        """
        if cls.__fec_data._simulation_time_step_per_ms is None:
            raise cls._exception("simulation_time_step_per_ms")
        return cls.__fec_data._simulation_time_step_per_ms

    @classmethod
    def get_simulation_time_step_per_s(cls):
        """
        The number of simulation time steps in a seconds.

        Syntactic sugar for `1,000,000 / simulation_time_step_us`

        :rtype: float
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the simulation_time_step is currently unavailable
        """
        if cls.__fec_data._simulation_time_step_per_s is None:
            raise cls._exception("simulation_time_step_per_s")
        return cls.__fec_data._simulation_time_step_per_s

    @classmethod
    def get_hardware_time_step_ms(cls):
        """
        The hardware timestep, in milliseconds.

        Syntactic sugar for `simulation_time_step_ms * time_scale_factor`

        :rtype: float
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the hardware_time_step is currently unavailable
        """
        if cls.__fec_data._hardware_time_step_ms is None:
            raise cls._exception("hardware_time_step_ms")
        return cls.__fec_data._hardware_time_step_ms

    @classmethod
    def get_hardware_time_step_us(cls):
        """
        The hardware timestep, in microseconds.

        Syntactic sugar for `simulation_time_step_us * time_scale_factor`

        :rtype: int
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the hardware_time_step is currently unavailable
        """
        if cls.__fec_data._hardware_time_step_us is None:
            raise cls._exception("ardware_time_step_us")
        return cls.__fec_data._hardware_time_step_us

    # time scale factor

    @classmethod
    def get_time_scale_factor(cls):
        """
        :rtype: int or float
        :raises SpiNNUtilsException:
            If the time_scale_factor is currently unavailable
        :raises SpiNNUtilsException:
            If the time_scale_factor is currently unavailable
        """
        if cls.__fec_data._time_scale_factor is None:
            raise cls._exception("time_scale_factor")
        return cls.__fec_data._time_scale_factor

    @classmethod
    def has_time_scale_factor(cls):
        """
        :rtype: bool
        """
        return cls.__fec_data._time_scale_factor is not None

    #  reset number

    @classmethod
    def get_reset_number(cls):
        """
        Get the number of times a reset has happened.

        Only counts the first reset after each run.

        So resets that are first soft then hard are ignored.
        Double reset calls without a run and resets before run are ignored.

        Reset numbers start at zero

        :rtype: int
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the run_number is currently unavailable
        """
        if cls.__fec_data._reset_number is None:
            raise cls._exception("run_number")
        return cls.__fec_data._reset_number

    @classmethod
    def get_reset_str(cls):
        """
        Get the number of times a reset has happened as a string.

        An empty string is returned if the system has never been reset
        (i.e., the reset number is 0)

        Only counts the first reset after each run.

        So resets that are first soft then hard are ignored.
        Double reset calls without a run and resets before run are ignored.

        Reset numbers start at zero

        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the run_number is currently unavailable
        :rtype: str
        """
        if cls.__fec_data._reset_number is None:
            raise cls._exception("run_number")
        if cls.__fec_data._reset_number:
            return str(cls.__fec_data._reset_number)
        else:
            return ""

    #  run number

    @classmethod
    def get_run_number(cls):
        """
        Get the number of this or the next run.

        Run numbers start at 1

        :rtype: int
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the run_number is currently unavailable
        """
        if cls.__fec_data._run_number is None:
            raise cls._exception("run_number")
        return cls.__fec_data._run_number

    @classmethod
    def get_run_step(cls):
        """
        Get the auto pause and resume step currently running if any.

        If and only if currently in an auto pause and resume loop this will
        report the number of the step. Starting at 1

        In most cases this will return `None`, including when running without
        steps.

        :rtype: None or int
        """
        return cls.__fec_data._run_step

    # Report directories
    # There are NO has or get methods for directories
    # This allows directories to be created on the fly

    # n_boards/chips required

    @classmethod
    def has_n_boards_required(cls):
        """
        Reports if a user has sets the number of boards requested during setup.

        :rtype: bool
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If n_boards_required is not set or set to `None`
        """
        return cls.__fec_data._n_boards_required is not None

    @classmethod
    def get_n_boards_required(cls):
        """
        Gets the number of boards requested by the user during setup if known.

        Guaranteed to be positive

        :rtype: int
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the n_boards_required is currently unavailable
        """
        if cls.__fec_data._n_boards_required is None:
            raise cls._exception("n_boards_requiredr")
        return cls.__fec_data._n_boards_required

    @classmethod
    def get_n_chips_needed(cls):
        """
        Gets the number of chips needed, if set.

        This will be the number of chips requested by the user during setup,
        even if this is less that what the partitioner reported.

        If the partitioner has run and the user has not specified a number,
        this will be what the partitioner requested.

        Guaranteed to be positive if set

        :rtype: int
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If data for n_chips_needed is not available
        """
        if cls.__fec_data._n_chips_required:
            return cls.__fec_data._n_chips_required
        if cls.__fec_data._n_chips_in_graph:
            return cls.__fec_data._n_chips_in_graph
        raise cls._exception("n_chips_requiredr")

    @classmethod
    def has_n_chips_needed(cls):
        """
        Detects if the number of chips needed has been set.

        This will be the number of chips requested by the use during setup or
        what the partitioner requested.

        :rtype: bool
        """
        if cls.__fec_data._n_chips_required is not None:
            return True
        return cls.__fec_data._n_chips_in_graph is not None

    @classmethod
    def get_timestamp_dir_path(cls):
        """
        Returns path to existing time-stamped directory in the reports
        directory.

        .. note::
            In unit-test mode this returns a temporary directory
            shared by all path methods.

        :rtype: str
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the simulation_time_step is currently unavailable
        """
        if cls.__fec_data._timestamp_dir_path:
            return cls.__fec_data._timestamp_dir_path
        if cls._is_mocked():
            return cls._temporary_dir_path()
        raise cls._exception("timestamp_dir_path")

    # system multicast routing data

    @classmethod
    def get_data_in_multicast_key_to_chip_map(cls):
        """
        Retrieve the data_in_multicast_key_to_chip_map if known.

        :rtype: dict
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the data_in_multicast_key_to_chip_map is currently unavailable
        """
        if cls.__fec_data._data_in_multicast_key_to_chip_map is None:
            raise cls._exception("data_in_multicast_key_to_chip_map")
        return cls.__fec_data._data_in_multicast_key_to_chip_map

    @classmethod
    def get_data_in_multicast_routing_tables(cls):
        """
        Retrieve the data_in_multicast_routing_tables if known

        :rtype: ~pacman.model.routing_tables.MulticastRoutingTables
        :raises SpiNNUtilsException:
            If the data_in_multicast_routing_tables is currently unavailable
        """
        if cls.__fec_data._data_in_multicast_routing_tables is None:
            raise cls._exception("data_in_multicast_routing_tables")
        return cls.__fec_data._data_in_multicast_routing_tables

    @classmethod
    def get_system_multicast_router_timeout_keys(cls):
        """
        Retrieve the system_multicast_router_timeout_keys if known.

        :rtype: dict
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the system_multicast_router_timeout_keys is currently
            unavailable
        """
        if cls.__fec_data._system_multicast_router_timeout_keys is None:
            raise cls._exception("system_multicast_router_timeout_keys")
        return cls.__fec_data._system_multicast_router_timeout_keys

    # ipaddress

    @classmethod
    def has_ipaddress(cls):
        """
        Detects if the IP address of the board with chip 0,0 is known.

        :rtype: bool
        """
        return cls.__fec_data._ipaddress is not None

    @classmethod
    def get_ipaddress(cls):
        """
        Gets the IP address of the board with chip 0,0 if it has been set.

        :rtype: str
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the IP address is currently unavailable
        """
        if cls.__fec_data._ipaddress is None:
            if cls._is_mocked():
                return "127.0.0.1"
            raise cls._exception("ipaddress")
        return cls.__fec_data._ipaddress

    # fixed_routes
    @classmethod
    def get_fixed_routes(cls):
        """
        Gets the fixed routes if they have been created.

        :rtype: dict(tuple(int,int), ~spinn_machine.FixedRouteEntry)
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the fixed_routes is currently unavailable
        """
        if cls.__fec_data._fixed_routes is None:
            raise cls._exception("fixed_routes")
        return cls.__fec_data._fixed_routes

    @classmethod
    def has_java_caller(cls):
        """
        Reports if there is a Java called that can be used.

        Equivalent to `get_config_bool("Java", "use_java")` as the writer will
        have created the caller during setup

        The behaviour when Mocked is currently to always return False.

        :rtype: bool
        """
        return cls.__fec_data._java_caller is not None

    @classmethod
    def get_java_caller(cls):
        """
        Gets the Java_caller.

        :rtype: str
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the java_caller is currently unavailable
        """
        if cls.__fec_data._java_caller is None:
            raise cls._exception("java_caller")
        return cls.__fec_data._java_caller

    # run_dir_path in UtilsDataView

    @classmethod
    def get_json_dir_path(cls):
        """
        Returns the path to the directory that holds all JSON files.

        This will be the path used by the last run call or to be used by
        the next run if it has not yet been called.

        .. note::
            In unit-test mode this returns a temporary directory
            shared by all path methods.

        :rtype: str
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the simulation_time_step is currently unavailable
        """
        if cls._is_mocked():
            return cls._temporary_dir_path()

        return cls._child_folder(cls.get_run_dir_path(), "json_files")

    @classmethod
    def get_provenance_dir_path(cls):
        """
        Returns the path to the directory that holds all provenance files.

        This will be the path used by the last run call or to be used by
        the next run if it has not yet been called.

        .. note::
            In unit-test mode this returns a temporary directory
            shared by all path methods.

        :rtype: str
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the simulation_time_step is currently unavailable
        """
        if cls._is_mocked():
            return cls._temporary_dir_path()
        return cls._child_folder(cls.get_run_dir_path(), "provenance_data")

    @classmethod
    def get_app_provenance_dir_path(cls):
        """
        Returns the path to the directory that holds all application provenance
        files.

        This will be the path used by the last run call or to be used by
        the next run if it has not yet been called.

        .. note::
            In unit-test mode this returns a temporary directory
            shared by all path methods.

        :rtype: str
        :raises ~spinn_utilities.exceptions.SimulatorNotSetupException:
            If the simulator has not been setup
        """
        if cls._is_mocked():
            return cls._temporary_dir_path()

        return cls._child_folder(
            cls.get_provenance_dir_path(), "app_provenance_data")

    @classmethod
    def get_system_provenance_dir_path(cls):
        """
        Returns the path to the directory that holds system provenance files.

        This will be the path used by the last run call or to be used by
        the next run if it has not yet been called.

        .. note::
            In unit-test mode this returns a temporary directory
            shared by all path methods.

        :rtype: str
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the simulation_time_step is currently unavailable
        """
        if cls._is_mocked():
            return cls._temporary_dir_path()
        return cls._child_folder(
            cls.get_provenance_dir_path(), "system_provenance_data")

    @classmethod
    def _child_folder(cls, parent, child_name, must_create=False):
        """
        :param str parent:
        :param str child_name:
        :param bool must_create:
            If `True`, the directory named by `child_name` (but not necessarily
            its parents) must be created by this call, and an exception will be
            thrown if this fails.
        :return: The fully qualified name of the child folder.
        :rtype: str
        :raises OSError:
            If the directory existed ahead of time and creation
            was required by the user
        """
        child = os.path.join(parent, child_name)
        if must_create:
            # Throws OSError or FileExistsError (a subclass of OSError) if the
            # directory exists.
            os.makedirs(child)
        elif not os.path.exists(child):
            os.makedirs(child, exist_ok=True)
        return child

    @classmethod
    def get_next_none_labelled_edge_number(cls):
        """
        Returns an unused number for labelling an unlabelled edge.

        :rtype: int
        """
        cls.__fec_data._none_labelled_edge_count += 1
        return cls.__fec_data._none_labelled_edge_count

    @classmethod
    def get_next_sync_signal(cls):
        """
        Returns alternately Signal.SYNC0 and Signal.SYNC1.

        :rtype: ~spinnman.messages.scp.enums.Signal
        """
        if cls.__fec_data._next_sync_signal == Signal.SYNC0:
            cls.__fec_data._next_sync_signal = Signal.SYNC1
            return Signal.SYNC0
        else:
            cls.__fec_data._next_sync_signal = Signal.SYNC0
            return Signal.SYNC1

    @classmethod
    def get_executable_types(cls):
        """
        Gets the _executable_types if they have been created.

        :rtype: dict(
            ~spinn_front_end_common.utilities.utility_objs.ExecutableType,
            ~spinn_machine.CoreSubsets)
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the executable_types is currently unavailable
        """
        if cls.__fec_data._executable_types is None:
            raise cls._exception("executable_types")
        return cls.__fec_data._executable_types

    @classmethod
    def has_live_packet_recorder_params(cls):
        """
        Reports if there are live_packet_recorder_params.

        If True the live_packet_recorder_params not be empty

        :rtype: bool
        """
        return cls.__fec_data._live_packet_recorder_params is not None

    @classmethod
    def get_live_packet_recorder_params(cls):
        """
        Mapping of live_packet_gatherer_params to a list of tuples
        (vertex and list of ids)).

        :rtype: dict(LivePacketGatherParameters,
            tuple(~pacman.model.graphs.AbstractVertex, list(str))
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the _live_packet_recorder_params is currently unavailable
        """
        if cls.__fec_data._live_packet_recorder_params is None:
            raise cls._exception("live_packet_recorder_params")
        return cls.__fec_data._live_packet_recorder_params

    # Add method in view so add can be done without going through ASB
    @classmethod
    def add_live_packet_gatherer_parameters(
            cls, live_packet_gatherer_params, vertex_to_record_from,
            partition_ids):
        """
        Adds parameters for a new LPG if needed, or adds to the tracker
        for parameters.

        .. note::
            If the
            :py:class:`~pacman.model.graphs.application.ApplicationGraph`
            is used, the vertex must be an
            :py:class:`~pacman.model.graphs.application.ApplicationVertex`.
            If not, it must be a
            :py:class:`~pacman.model.graphs.machine.MachineVertex`.

        :param LivePacketGatherParameters live_packet_gatherer_params:
            parameters for an LPG to look for or create
        :param ~pacman.model.graphs.AbstractVertex vertex_to_record_from:
            the vertex that needs to send to a given LPG
        :param iterable(str) partition_ids:
            the IDs of the partitions to connect from the vertex
        """
        if cls.__fec_data._live_packet_recorder_params is None:
            # pylint: disable=attribute-defined-outside-init
            cls.__fec_data._live_packet_recorder_params = dict()
        lpg_vertex = cls.__fec_data._live_packet_recorder_params.get(
            live_packet_gatherer_params)
        if lpg_vertex is None:
            # UGLY import due to circular reference
            from spinn_front_end_common.utility_models import LivePacketGather
            lpg_vertex = LivePacketGather(
                live_packet_gatherer_params, live_packet_gatherer_params.label)
            cls.__fec_data._live_packet_recorder_params[
                live_packet_gatherer_params] = lpg_vertex
            cls.add_vertex(lpg_vertex)
        for part_id in partition_ids:
            cls.add_edge(
                ApplicationEdge(vertex_to_record_from, lpg_vertex), part_id)

    @classmethod
    def get_database_file_path(cls):
        """
        Will return the database_file_path if set or `None` if not set
        or set to `None`

        :rtype: str or None
        """
        return cls.__fec_data._database_file_path

    @classmethod
    def get_executable_targets(cls):
        """
        Binaries to be executed.

        :rtype: ~spinnman.model.ExecutableTargets
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the executable_targets is currently unavailable
        """
        if cls.__fec_data._executable_targets is None:
            raise cls._exception("executable_targets")
        return cls.__fec_data._executable_targets

    @classmethod
    def get_dsg_targets(cls):
        """
        Data Spec targets database.

        :rtype: DsSqlliteDatabase
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the dsg_targets is currently unavailable
        """
        if cls.__fec_data._dsg_targets is None:
            raise cls._exception("dsg_targets")
        return cls.__fec_data._dsg_targets

    @classmethod
    def has_monitors(cls):
        """
        Detect is ExtraMonitorSupportMachineVertex(s) have been created.

        :rtype: bool
        """
        return cls.__fec_data._monitor_map is not None

    @classmethod
    def get_monitor_by_xy(cls, x, y):
        """
        The ExtraMonitorSupportMachineVertex for chip (x,y).

        :param int x: X coordinate of chip
        :param int y: Y coordinate of chip
        :rtype: ExtraMonitorSupportMachineVertex
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the monitors are currently unavailable
        :raises KeyError: If chip (x,y) does not have a monitor
        """
        if cls.__fec_data._monitor_map is None:
            raise cls._exception("monitors_map")
        # pylint: disable=unsubscriptable-object
        return cls.__fec_data._monitor_map[(x, y)]

    @classmethod
    def iterate_monitor_items(cls):
        """
        Iterates over the (x,y) and ExtraMonitorSupportMachineVertex.

        get_n_monitors returns the number of items this iterable will provide.

        :rtype: iterable(tuple(tuple(int,int),
            ExtraMonitorSupportMachineVertex))
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the monitors are currently unavailable
        """
        if cls.__fec_data._monitor_map is None:
            raise cls._exception("monitors_map")
        return cls.__fec_data._monitor_map.items()

    @classmethod
    def get_n_monitors(cls):
        """
        Number of ExtraMonitorSupportMachineVertexs.

        :rtype: int
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the monitors are currently unavailable
        """
        if cls.__fec_data._monitor_map is None:
            raise cls._exception("monitors_map")
        return len(cls.__fec_data._monitor_map)

    @classmethod
    def iterate_monitors(cls):
        """
        Iterates over the ExtraMonitorSupportMachineVertex(s).

        :rtype: iterable(ExtraMonitorSupportMachineVertex)
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the monitors are currently unavailable
        """
        if cls.__fec_data._monitor_map is None:
            raise cls._exception("monitors_map")
        return cls.__fec_data._monitor_map.values()

    @classmethod
    def get_gatherer_by_xy(cls, x, y):
        """
        The DataSpeedUpPacketGatherMachineVertex for chip (x,y).

        :param int x: X coordinate of chip
        :param int y: Y coordinate of chip
        :rtype: DataSpeedUpPacketGatherMachineVertex
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the gatherers are currently unavailable
        :raises KeyError: If chip (x,y) does not have a monitor
        """
        if cls.__fec_data._gatherer_map is None:
            raise cls._exception("gatherer_map")
        # pylint: disable=unsubscriptable-object
        return cls.__fec_data._gatherer_map[(x, y)]

    @classmethod
    def iterate_gather_items(cls):
        """
        Iterates over the (x,y) and DataSpeedUpPacketGatherMachineVertex.

        get_n_gathers returns the number of items this iterable will provide

        :rtype: iterable(tuple(tuple(int,int),
             DataSpeedUpPacketGatherMachineVertex))
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the gathers are currently unavailable
        """
        if cls.__fec_data._gatherer_map is None:
            raise cls._exception("gatherer_map")
        return cls.__fec_data._gatherer_map.items()

    @classmethod
    def get_n_gathers(cls):
        """
        Number of DataSpeedUpPacketGatherMachineVertex(s).

        :rtype: int
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the gathers are currently unavailable
        """
        if cls.__fec_data._gatherer_map is None:
            raise cls._exception("gatherer_map")
        return len(cls.__fec_data._gatherer_map)

    @classmethod
    def iterate_gathers(cls):
        """
        Iterates over the DataSpeedUpPacketGatherMachineVertex(s).

        :rtype: iterable(DataSpeedUpPacketGatherMachineVertex)
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the gathers are currently unavailable
        """
        if cls.__fec_data._gatherer_map is None:
            raise cls._exception("gatherer_map")
        return cls.__fec_data._gatherer_map.values()

    @classmethod
    def iterate_database_socket_addresses(cls):
        """
        Iterates over the registered database_socket_addresses.

        :rtype: iterable(~spinn_utilities.socket_address.SocketAddress)
        """
        return iter(cls.__fec_data._database_socket_addresses)

    @classmethod
    def get_n_database_socket_addresses(cls):
        """
        Number of registered database_socket_addresses.

        :rtype: int
        """
        return len(cls.__fec_data._database_socket_addresses)

    @classmethod
    def add_database_socket_address(cls, database_socket_address):
        """
        Adds a socket address to the list of known addresses.

        :param database_socket_address:
        :type database_socket_address:
            ~spinn_utilities.socket_address.SocketAddress
        :raises TypeError: if database_socket_address is not a SocketAddress
        """
        if not isinstance(database_socket_address, SocketAddress):
            raise TypeError("database_socket_address must be a SocketAddress")
        cls.__fec_data._database_socket_addresses.add(database_socket_address)

    @classmethod
    def add_database_socket_addresses(cls, database_socket_addresses):
        """
        Adds all socket addresses to the list of known addresses.

        :param database_socket_addresses: The addresses to add
        :type database_socket_addresses:
            iterable(~spinn_utilities.socket_address.SocketAddress)
        :raises TypeError:
           if database_socket_address is not a iterable of `SocketAddress`
        """
        if database_socket_addresses is None:
            return
        for socket_address in database_socket_addresses:
            cls.add_database_socket_address(socket_address)

    @classmethod
    def get_notification_protocol(cls):
        """
        The notification protocol handler.

        :rtype: NotificationProtocol
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the notification_protocol is currently unavailable
        """
        if cls.__fec_data._notification_protocol is None:
            raise cls._exception("notification_protocol")
        return cls.__fec_data._notification_protocol

    @classmethod
    def add_live_output_vertex(cls, vertex, partition_id):
        """
        Add a vertex that is to be output live, and so wants its atom IDs
        recorded in the database.

        :param ~pacman.model.graphs.application.ApplicationVertex vertex:
            The vertex to add
        :param str partition_id: The partition to get the IDs of
        """
        cls.__fec_data._live_output_vertices.add((vertex, partition_id))

    @classmethod
    def iterate_live_output_vertices(cls):
        """
        Get an iterator over the live output vertices and partition IDs.

        :rtype:
            iterable(tuple(~pacman.model.graphs.application.ApplicationVertex,
            str))
        """
        return iter(cls.__fec_data._live_output_vertices)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from data_specification.constants import APP_PTR_TABLE_BYTE_SIZE

# conversion from words to bytes
BYTES_PER_WORD = 4
BYTES_PER_4_WORDS = 16
BYTES_PER_SHORT = 2
BYTES_PER_KB = 1024

LIVE_GATHERER_CORE_APPLICATION_ID = 0xAC0
COMMAND_SENDER_CORE_APPLICATION_ID = 0xAC6
SPIKE_INJECTOR_CORE_APPLICATION_ID = 0xAC9

# how many bits there are in a word
BITS_PER_WORD = 32.0

#: start of where SDRAM starts (either unbuffered or buffered)
SDRAM_BASE_ADDR = 0x70000000

#: the ITCM max safe limit for a binary
MAX_SAFE_BINARY_SIZE = 32 * BYTES_PER_KB

#: the ITCM max limit for a binary
MAX_POSSIBLE_BINARY_SIZE = 33 * BYTES_PER_KB

# converts between micro and milli seconds
MICRO_TO_MILLISECOND_CONVERSION = 1000.0
MICRO_TO_SECOND_CONVERSION = 1000000.0  # (1e6)

#: max size expected to be used by the reverse ip_tag multicast source
#: during buffered operations.
MAX_SIZE_OF_BUFFERED_REGION_ON_CHIP = 1 * 1024 * BYTES_PER_KB

#: The default size of a recording buffer before receive request is sent
DEFAULT_BUFFER_SIZE_BEFORE_RECEIVE = 16 * BYTES_PER_KB

#: The number of bytes used by SARK per memory allocation
SARK_PER_MALLOC_SDRAM_USAGE = 2 * BYTES_PER_WORD

#: The number of words in the AbstractDataSpecable basic setup information.
#: This is the amount required by the pointer table plus a SARK allocation.
DATA_SPECABLE_BASIC_SETUP_INFO_N_BYTES = (
    APP_PTR_TABLE_BYTE_SIZE + SARK_PER_MALLOC_SDRAM_USAGE)

#: The number of bytes used by the simulation interface.
#: This is one word for the machine_time_step, one for the SDP port, and one
#: for the application hash.
SIMULATION_N_BYTES = 3 * BYTES_PER_WORD

#: the number of bytes used by the multicast data speed up interface
# 4 for the first key used by multicast protocol
MULTICAST_SPEEDUP_N_BYTES = BYTES_PER_WORD

#: The number of bytes used by the DSG and simulation interfaces
SYSTEM_BYTES_REQUIREMENT = (
    DATA_SPECABLE_BASIC_SETUP_INFO_N_BYTES + SIMULATION_N_BYTES)

#: Database file path maximum length for database notification messages.
#: Note that this is *not* sent to SpiNNaker and so is not subject to the
#: usual SDP limit.
MAX_DATABASE_PATH_LENGTH = 50000

#: size of the on-chip DSE data structure required, in bytes
DSE_DATA_STRUCT_SIZE = 4 * BYTES_PER_WORD


class SDP_RUNNING_MESSAGE_CODES(Enum):
    """
    Codes for sending control messages to spin1_api.
    """
    SDP_STOP_ID_CODE = 6
    SDP_NEW_RUNTIME_ID_CODE = 7
    SDP_UPDATE_PROVENCE_REGION_AND_EXIT = 8
    SDP_CLEAR_IOBUF_CODE = 9


class SDP_PORTS(Enum):
    """
    SDP port handling output buffering data streaming.
    """

    #: Command port for the buffered in functionality.
    INPUT_BUFFERING_SDP_PORT = 1
    #: Command port for the buffered out functionality.
    OUTPUT_BUFFERING_SDP_PORT = 2
    #: Command port for resetting runtime, etc.
    #: See :py:class:`SDP_RUNNING_MESSAGE_CODES`
    RUNNING_COMMAND_SDP_PORT = 3
    #: Extra monitor core reinjection control protocol.
    #: See :py:class:`ReinjectorSCPCommands`
    EXTRA_MONITOR_CORE_REINJECTION = 4
    #: Extra monitor core outbound data transfer protocol
    EXTRA_MONITOR_CORE_DATA_SPEED_UP = 5
    #: Extra monitor core inbound data transfer protocol
    #: See :py:class:`SpeedupInSCPCommands`
    EXTRA_MONITOR_CORE_DATA_IN_SPEED_UP = 6


# output buffering operations
class BUFFERING_OPERATIONS(Enum):
    """
    A listing of what SpiNNaker specific EIEIO commands there are.
    """

    #: Database handshake with external program
    BUFFER_READ = 0
    #: Host confirming data being read form SpiNNaker memory
    BUFFER_WRITE = 1


#: partition IDs preallocated to functionality
PARTITION_ID_FOR_MULTICAST_DATA_SPEED_UP = "DATA_SPEED_UP_ROAD"

#: The default local port that the toolchain listens on for the notification
#: protocol.
NOTIFY_PORT = 19999

#: The number of clock cycles per micro-second (at 200Mhz)
CLOCKS_PER_US = 200

PROVENANCE_DB = "provenance.sqlite3"

#: SDRAM Tag used by the compressor to find the routing tables
COMPRESSOR_SDRAM_TAG = 1

#: SDRAM Tags used for bitfield compressor
BIT_FIELD_COMMS_SDRAM_TAG = 2
BIT_FIELD_USABLE_SDRAM_TAG = 3
BIT_FIELD_ADDRESSES_SDRAM_TAG = 4
BIT_FIELD_ROUTING_TABLE_SDRAM_TAG = 5

#: Base SDRAM tag used by SDRAM edges when allocating
#: (allows up to 100 edges per chip)
SDRAM_EDGE_BASE_TAG = 100

#: Base SDRAM tag used by cores when loading data
#: (tags 201-217 will be used by cores 1-17)
CORE_DATA_SDRAM_BASE_TAG = 200



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.data import FecDataView


def fixed_route_from_machine_report():
    """
    Writes the fixed routes from the machine.
    """
    file_name = os.path.join(
        FecDataView.get_run_dir_path(), "fixed_route_routers")
    transceiver = FecDataView.get_transceiver()
    machine = FecDataView.get_machine()

    progress = ProgressBar(machine.n_chips, "Writing fixed route report")

    app_id = FecDataView.get_app_id()
    with open(file_name, "w", encoding="utf-8") as f:
        f.write(" x    y       route         [cores][links]\n")
        for chip in progress.over(machine.chips):
            fixed_route = transceiver.read_fixed_route(
                chip.x, chip.y, app_id)
            f.write("{: <3s}:{: <3s} contains route {: <10s} {}\n".format(
                str(chip.x), str(chip.y),
                _reduce_route_value(
                    fixed_route.processor_ids, fixed_route.link_ids),
                _expand_route_value(
                    fixed_route.processor_ids, fixed_route.link_ids)))


def _reduce_route_value(processors_ids, link_ids):
    value = 0
    for link in link_ids:
        value += 1 << link
    for processor in processors_ids:
        value += 1 << (processor + 6)
    return str(value)


def _expand_route_value(processors_ids, link_ids):
    """
    Convert a 32-bit route word into a string which lists the target
    cores and links.
    """

    # Convert processor targets to readable values:
    route_string = "["
    separator = ""
    for processor in processors_ids:
        route_string += separator + str(processor)
        separator = ", "

    route_string += "] ["

    # Convert link targets to readable values:
    link_labels = {0: 'E', 1: 'NE', 2: 'N', 3: 'W', 4: 'SW', 5: 'S'}

    separator = ""
    for link in link_ids:
        route_string += separator + link_labels[link]
        separator = ", "
    route_string += "]"
    return route_string



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import numpy

from pacman.model.partitioner_interfaces import LegacyPartitionerAPI
from spinn_utilities.overrides import overrides
from pacman.model.graphs.application import ApplicationVertex
from spinn_front_end_common.utilities.constants import SDP_PORTS
from .reverse_ip_tag_multicast_source_machine_vertex import (
    ReverseIPTagMulticastSourceMachineVertex)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from pacman.model.routing_info.base_key_and_mask import BaseKeyAndMask


class ReverseIpTagMultiCastSource(
        ApplicationVertex, LegacyPartitionerAPI):
    """
    A model which will allow events to be injected into a SpiNNaker
    machine and converted into multicast packets.
    """

    def __init__(
            self, n_keys, label=None, max_atoms_per_core=sys.maxsize,

            # Live input parameters
            receive_port=None,
            receive_sdp_port=SDP_PORTS.INPUT_BUFFERING_SDP_PORT.value,
            receive_tag=None,
            receive_rate=10,

            # Key parameters
            virtual_key=None, prefix=None,
            prefix_type=None, check_keys=False,

            # Send buffer parameters
            send_buffer_times=None,
            send_buffer_partition_id=None,

            # Extra flag for input without a reserved port
            reserve_reverse_ip_tag=False,

            # Name of partition to inject keys with
            injection_partition_id=None,

            # splitter object
            splitter=None):
        """
        :param int n_keys:
            The number of keys to be sent via this multicast source
        :param str label: The label of this vertex
        :param int max_atoms_per_core:
        :param board_address: The IP address of the board on which to place
            this vertex if receiving data, either buffered or live (by
            default, any board is chosen)
        :type board_address: str or None
        :param receive_port: The port on the board that will listen for
            incoming event packets (default is to disable this feature; set a
            value to enable it)
        :type receive_port: int or None
        :param int receive_sdp_port:
            The SDP port to listen on for incoming event packets
            (defaults to 1)
        :param ~spinn_machine.tags.IPTag receive_tag:
            The IP tag to use for receiving live events
            (uses any by default)
        :param float receive_rate:
            The estimated rate of packets that will be sent by this source
        :param int virtual_key:
            The base multicast key to send received events with
            (assigned automatically by default)
        :param int prefix:
            The prefix to "or" with generated multicast keys
            (default is no prefix)
        :param ~spinnman.messages.eieio.EIEIOPrefix prefix_type:
            Whether the prefix should apply to the upper or lower half of the
            multicast keys (default is upper half)
        :param bool check_keys:
            True if the keys of received events should be verified before
            sending (default False)
        :param send_buffer_times: An array of arrays of times at which keys
            should be sent (one array for each key, default disabled)
        :type send_buffer_times:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(~numpy.int32)) or None
        :param send_buffer_partition_id: The ID of the partition containing
            the edges down which the events are to be sent
        :type send_buffer_partition_id: str or None
        :param bool reserve_reverse_ip_tag:
            Extra flag for input without a reserved port
        :param str injection_partition:
            If not `None`, will enable injection and specify the partition to
            send injected keys with
        :param splitter: the splitter object needed for this vertex
        :type splitter: None or
            ~pacman.model.partitioner_splitters.abstract_splitters.AbstractSplitterCommon
        """
        # pylint: disable=too-many-arguments
        super().__init__(label, max_atoms_per_core, splitter=splitter)

        # basic items
        self._n_atoms = self.round_n_atoms(n_keys, "n_keys")

        # Store the parameters for EIEIO
        self._receive_port = receive_port
        self._receive_sdp_port = receive_sdp_port
        self._receive_tag = receive_tag
        self._receive_rate = receive_rate
        self._virtual_key = virtual_key
        self._prefix = prefix
        self._prefix_type = prefix_type
        self._check_keys = check_keys

        # Store the send buffering details
        self._send_buffer_times = self._validate_send_buffer_times(
            send_buffer_times)
        self._send_buffer_partition_id = send_buffer_partition_id

        # Store the buffering details
        self._reserve_reverse_ip_tag = reserve_reverse_ip_tag
        self._injection_partition_id = injection_partition_id

        # Store recording parameters
        self._is_recording = False

    def _validate_send_buffer_times(self, send_buffer_times):
        if send_buffer_times is None:
            return None
        if len(send_buffer_times) and hasattr(send_buffer_times[0], "__len__"):
            if len(send_buffer_times) != self._n_atoms:
                raise ConfigurationException(
                    f"The array or arrays of times {send_buffer_times} does "
                    f"not have the expected length of {self._n_atoms}")
            return numpy.array(send_buffer_times, dtype="object")
        return numpy.array(send_buffer_times)

    @property
    @overrides(ApplicationVertex.n_atoms)
    def n_atoms(self):
        return self._n_atoms

    @overrides(LegacyPartitionerAPI.get_sdram_used_by_atoms)
    def get_sdram_used_by_atoms(self, vertex_slice):
        return ReverseIPTagMulticastSourceMachineVertex.get_sdram_usage(
            self._filtered_send_buffer_times(vertex_slice),
            self._is_recording, self._receive_rate, vertex_slice.n_atoms)

    @property
    def send_buffer_times(self):
        """
        When messages will be sent.

        :rtype: ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(~numpy.int32)) or None
        """
        return self._send_buffer_times

    @send_buffer_times.setter
    def send_buffer_times(self, send_buffer_times):
        self._send_buffer_times = send_buffer_times
        for vertex in self.machine_vertices:
            send_buffer_times_to_set = self._send_buffer_times
            if len(self._send_buffer_times) > 0:
                if hasattr(self._send_buffer_times[0], "__len__"):
                    vertex_slice = vertex.vertex_slice
                    send_buffer_times_to_set = self._send_buffer_times[
                        vertex_slice.lo_atom:vertex_slice.hi_atom + 1]
            vertex.send_buffer_times = send_buffer_times_to_set

    def enable_recording(self, new_state=True):
        self._is_recording = new_state

    @overrides(LegacyPartitionerAPI.create_machine_vertex)
    def create_machine_vertex(
            self, vertex_slice, sdram, label=None):
        send_buffer_times = self._filtered_send_buffer_times(vertex_slice)
        machine_vertex = ReverseIPTagMulticastSourceMachineVertex(
            vertex_slice=vertex_slice,
            label=label, app_vertex=self,
            receive_port=self._receive_port,
            receive_sdp_port=self._receive_sdp_port,
            receive_tag=self._receive_tag,
            receive_rate=self._receive_rate,
            virtual_key=self._virtual_key, prefix=self._prefix,
            prefix_type=self._prefix_type, check_keys=self._check_keys,
            send_buffer_times=send_buffer_times,
            send_buffer_partition_id=self._send_buffer_partition_id,
            reserve_reverse_ip_tag=self._reserve_reverse_ip_tag,
            injection_partition_id=self._injection_partition_id)
        machine_vertex.enable_recording(self._is_recording)
        # Known issue with ReverseIPTagMulticastSourceMachineVertex
        if sdram:
            assert (sdram == machine_vertex.sdram_required)
        return machine_vertex

    def _filtered_send_buffer_times(self, vertex_slice):
        ids = vertex_slice.get_raster_ids(self.atoms_shape)
        send_buffer_times = self._send_buffer_times
        n_buffer_times = 0
        if send_buffer_times is not None:
            # If there is at least one array element, and that element is
            # itself an array
            if (len(send_buffer_times) and
                    hasattr(send_buffer_times[0], "__len__")):
                send_buffer_times = send_buffer_times[ids]
            # Check the buffer times are not empty
            for i in send_buffer_times:
                if hasattr(i, "__len__"):
                    n_buffer_times += len(i)
                else:
                    # assuming this must be a single integer
                    n_buffer_times += 1
        if n_buffer_times == 0:
            return None
        return send_buffer_times

    def __repr__(self):
        return self._label

    @overrides(ApplicationVertex.get_fixed_key_and_mask)
    def get_fixed_key_and_mask(self, partition_id):
        if self._virtual_key is None:
            return None
        mask = ReverseIPTagMulticastSourceMachineVertex.calculate_mask(
            min(self.n_atoms, self.get_max_atoms_per_core()))
        return BaseKeyAndMask(self._virtual_key, mask)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_machine import CoreSubsets
from spinn_front_end_common.abstract_models import AbstractHasAssociatedBinary
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.utility_objs import ExecutableType


def locate_executable_start_type():
    """
    Discovers where applications of particular types need to be launched.

    """
    binary_start_types = dict()

    for placement in FecDataView.iterate_placements_by_vertex_type(
            AbstractHasAssociatedBinary):
        bin_type = placement.vertex.get_binary_start_type()
        # update core subset with location of the vertex on the
        # machine
        if bin_type not in binary_start_types:
            binary_start_types[bin_type] = CoreSubsets()

        __add_vertex_to_subset(
            placement, binary_start_types[bin_type])

    # only got apps with no binary, such as external devices.
    # return no app
    if not binary_start_types:
        return {ExecutableType.NO_APPLICATION: None}

    return binary_start_types


def __add_vertex_to_subset(placement, core_subsets):
    """
    :param ~.Placement placement:
    :param ~.CoreSubsets core_subsets:
    """
    core_subsets.add_processor(
        x=placement.x, y=placement.y, processor_id=placement.p)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractproperty
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.machine import MachineVertex


@require_subclass(MachineVertex)
class AbstractSupportsDatabaseInjection(object, metaclass=AbstractBase):
    """
    Marks a machine vertex as supporting injection of information via a
    database running on the controlling host.
    """

    __slots__ = ()

    @abstractproperty
    def is_in_injection_mode(self):
        """
        Whether this vertex is actually in injection mode.

        :rtype: bool
        """

    @abstractproperty
    def injection_partition_id(self):
        """
        The partition that packets are being injected with.

        :rtype: str
        """



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.data import FecDataView
logger = FormatAdapter(logging.getLogger(__name__))

_FOLDER_NAME = "memory_map_from_processor_to_address_space"


def memory_map_on_host_report():
    """
    Report on memory usage.
    """
    file_name = os.path.join(FecDataView.get_run_dir_path(), _FOLDER_NAME)
    try:
        with open(file_name, "w", encoding="utf-8") as f:
            f.write("On host data specification executor\n")
            for key, start_address, memory_used, memory_written in \
                    FecDataView.get_dsg_targets().info_iteritems():
                f.write(
                    f"{key}: ('start_address': {start_address}, "
                    f"hex:{hex(start_address)}), "
                    f"'memory_used': {memory_used}, "
                    f"'memory_written': {memory_written} \n")
    except IOError:
        logger.exception("Generate_placement_reports: Can't open file"
                         " {} for writing.", file_name)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class DPRIFlags(Enum):
    """
    SCP Dropped Packet Reinjection (DPRI) packet type flags.
    """
    MULTICAST = 1
    POINT_TO_POINT = 2
    NEAREST_NEIGHBOUR = 4
    FIXED_ROUTE = 8

    def __new__(cls, value, doc=""):
        # pylint: disable=protected-access
        obj = object.__new__(cls)
        obj._value_ = value
        obj.__doc__ = doc
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import traceback
from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.interface.provenance import (
    AbstractProvidesProvenanceDataFromMachine)

logger = FormatAdapter(logging.getLogger(__name__))


def placements_provenance_gatherer(n_placements, placements):
    """
    Gets provenance information from placements.

    :param int n_placements: Number of placements to gather
    :param iterator(~pacman.model.placements.Placement) placements:
        The placements of the vertices to gather data form.
        May not be all placements so don't use View
    """
    errors = list()

    progress = ProgressBar(n_placements, "Getting provenance data")

    # retrieve provenance data from any cores that provide data
    for placement in progress.over(placements):
        _add_placement_provenance(placement, errors)
    if errors:
        logger.warning("Errors found during provenance gathering:")
        for error in errors:
            logger.warning("{}", error)


def _add_placement_provenance(placement, errors):
    """
    :param ~.Placement placement:
    :param list(str) errors:
    """
    # retrieve provenance data from any cores that provide data
    if isinstance(
            placement.vertex, AbstractProvidesProvenanceDataFromMachine):
        # get data
        try:
            placement.vertex.get_provenance_data_from_machine(placement)

        except Exception:  # pylint: disable=broad-except
            errors.append(traceback.format_exc())



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.progress_bar import ProgressBar
from spinnman.constants import ROUTER_REGISTER_REGISTERS
from spinnman.model import DiagnosticFilter
from spinnman.model.enums import (
    DiagnosticFilterDefaultRoutingStatus, DiagnosticFilterPacketType,
    DiagnosticFilterSource)
from spinn_front_end_common.data import FecDataView


def routing_setup():
    """
    Initialises the routers.

    .. note::
        This does not load any routes into them.

    :param ~spinnman.transceiver.Transceiver transceiver:
    """
    transceiver = FecDataView.get_transceiver()
    routing_tables = FecDataView.get_uncompressed().routing_tables
    progress = ProgressBar(len(routing_tables), "Preparing Routing Tables")

    # Clear the routing table for each router that needs to be set up
    # and set up the diagnostics
    for router_table in progress.over(routing_tables):
        transceiver.clear_multicast_routes(
            router_table.x, router_table.y)
        transceiver.clear_router_diagnostic_counters(
            router_table.x, router_table.y)

        # set the router diagnostic for user 3 to catch local default
        # routed packets. This can only occur when the source router
        # has no router entry, and therefore should be detected a bad
        # dropped packet.
        __set_router_diagnostic_filters(
            router_table.x, router_table.y, transceiver)


def __set_router_diagnostic_filters(x, y, transceiver):
    """
    :param int x:
    :param int y:
    :param ~.Transceiver transceiver:
    """
    transceiver.set_router_diagnostic_filter(
        x, y, ROUTER_REGISTER_REGISTERS.USER_3.value,
        DiagnosticFilter(
            enable_interrupt_on_counter_event=False,
            match_emergency_routing_status_to_incoming_packet=False,
            destinations=[],
            sources=[DiagnosticFilterSource.LOCAL],
            payload_statuses=[],
            default_routing_statuses=[
                DiagnosticFilterDefaultRoutingStatus.DEFAULT_ROUTED],
            emergency_routing_statuses=[],
            packet_types=[DiagnosticFilterPacketType.MULTICAST]))

    transceiver.set_router_diagnostic_filter(
        x, y, ROUTER_REGISTER_REGISTERS.USER_2.value,
        DiagnosticFilter(
            enable_interrupt_on_counter_event=False,
            match_emergency_routing_status_to_incoming_packet=False,
            destinations=[],
            sources=[DiagnosticFilterSource.NON_LOCAL],
            payload_statuses=[],
            default_routing_statuses=[
                DiagnosticFilterDefaultRoutingStatus.DEFAULT_ROUTED],
            emergency_routing_statuses=[],
            packet_types=[DiagnosticFilterPacketType.MULTICAST]))
